from dataclasses import dataclass
from typing import Optional

import numpy as np

import ray
from agox.candidates import Candidate
from agox.collectors.ABC_collector import CollectorBaseClass
from agox.environments import Environment
from agox.generators import Generator
from agox.samplers import Sampler
from agox.samplers.fixed_sampler import FixedSampler
from agox.utils.ray import RayPoolUser


@dataclass
class Walker:
    index: int
    generator_index: int
    generator_pool_id: str

    @property
    def name(self) -> str:
        return f"walker-{self.index}"

def remote_generate(generator: Generator, sampler: Sampler, environment: Environment, walker: Walker) -> list[Candidate]:
    candidates = generator(sampler, environment)
    if len(candidates) == 1:
        candidate = candidates[0]
        candidate.add_meta_information("walker_index", walker.index)
    else:
        candidate = None
    return candidate


@dataclass
class DynamicGeneratorInfo:
    attribute: str
    correlated: bool
    caps: list[float]


class ReplicaExchangeCollector(CollectorBaseClass, RayPoolUser):
    name = "RepliceExchangeCollector"

    """
    Parameters
    ----------
    """

    def __init__(
        self, generator_dict: dict[int, Generator], fallback_generator: Optional[Generator] = None, **kwargs
    ) -> None:
        # Get the kwargs that are relevant for RayPoolUser
        RayPoolUser.__init__(self)
        self.dynamic_generators = {}

        # Determine the unique generators and their pool keys.
        unique_generators = []
        walker_index_to_unique_generator = {}
        for walker_index, generator in generator_dict.items():
            if generator not in unique_generators:
                unique_generators.append(generator)
                pool_key = self.pool_add_module(generator)
            walker_index_to_unique_generator[walker_index] = (unique_generators.index(generator), pool_key)

        # Give the rest to the CollectorBaseClass.
        CollectorBaseClass.__init__(self, unique_generators, **kwargs)

        # Setup walkers
        self.walkers = []
        for walker_index, generator in generator_dict.items():
            generator_index, pool_key = walker_index_to_unique_generator[walker_index]
            assert self.generators[generator_index] is generator
            walker = Walker(walker_index, generator_index, pool_key)
            self.walkers.append(walker)

        # Fallback
        if fallback_generator is None:
            fallback_generator = self.get_fallback_generator()
        self.fallback_generator_id = self.pool_add_module(fallback_generator)

    def make_candidates(self) -> list[Candidate]:
        # Update dynamic generators, e.g. rattle amplitudes
        if len(self.dynamic_generators) != 0:
            self.update_generators()

        environment_id = ray.put(self.environment)
        args = []
        kwargs = []
        modules = []
        for walker in self.walkers:
            # Retrieve parent candidate and create a sampler
            parent = self.sampler.get_walker(walker.index)
            sampler = FixedSampler(parent)
            sampler_id = ray.put(sampler)

            args.append((sampler_id, environment_id, walker))
            kwargs.append({})

            # Determine which generator to use
            if sampler.has_none():
                modules.append([self.fallback_generator_id])
            else:
                modules.append([walker.generator_pool_id])

        # Generate in parallel using the pool.
        candidates = self.pool_map(remote_generate, modules, args, kwargs)
        candidates = [c for c in candidates if c is not None]

        return candidates

    def get_number_of_candidates(self) -> list[int]:
        return [self.sampler.sample_size]

    @classmethod
    def from_sampler(
        cls,
        sampler: Sampler,
        environment: Environment,
        amplitudes: np.array,
        generator_dict: dict[int, Generator] = None,
        random_generator_kwargs: dict | None = None,
        rattle_kwargs: dict | None = None,
        **kwargs,
    ):
        from agox.generators import RandomGenerator, RattleGenerator

        # Validate inputs
        if not len(amplitudes) == len(sampler.temperatures):
            raise ValueError("The number of amplitudes must match the number of temperatures in the sampler.")
        
        generator_dict = generator_dict or {}
        rattle_kwargs = rattle_kwargs or {}
        random_generator_kwargs = random_generator_kwargs or {}

        # Setup generators for each walker
        for walker_index in range(sampler.sample_size):
            if walker_index in generator_dict.keys():
                continue

            if walker_index == sampler.sample_size - 1:
                # Setup random generator
                generator = RandomGenerator(**environment.get_confinement(), **random_generator_kwargs)
            else:
                # Setup rattle generators
                n_atoms = len(environment.get_missing_indices())
                generator = RattleGenerator(
                    **environment.get_confinement(), 
                    rattle_amplitude=amplitudes[walker_index], 
                    n_rattle=n_atoms,
                    **rattle_kwargs
                )
            generator_dict[walker_index] = generator

        fallback_generator = RandomGenerator(**environment.get_confinement(), **random_generator_kwargs)
        return cls(generator_dict, sampler=sampler, environment=environment, order=1, fallback_generator=fallback_generator, **kwargs)

    def add_generator_update(
        self,
        walker_index: int,
        attribute: str,
        correlated: bool = False,
        min_val: float = 0.25,
        max_val: float = 2,
    ) -> None:

        found = False        
        for walker in self.walkers:
            if walker.index == walker_index:
                found = True
                break

        if not found:
            raise ValueError(f"Walker with index {walker_index} not found.")

        generator = self.generators[walker.generator_index]
        if hasattr(generator, attribute):
            temperature = self.sampler.temperatures[walker_index]
            generator.add_dynamic_attribute(attribute)
            # self.dynamic_generators[walker_index] = DynamicGeneratorInfo(
            #     attribute, correlated, [min_val * temperature, max_val * temperature]
            # )
            self.dynamic_generators[walker_index] = DynamicGeneratorInfo(
                attribute, correlated, [min_val, max_val]
            )


    def update_generators(self) -> None:
        for walker in self.walkers:
            if walker.index not in self.dynamic_generators:
                continue
            generator = self.generators[walker.generator_index]
            acceptance = self.sampler.tracker.get_acceptance_rate(
                walker.index, start=-10
            )  # Acceptance rate over the last 10 iterations.

            if acceptance.size == 0:
                continue

            info = self.dynamic_generators[walker.index]

            attribute = info.attribute
            caps = info.caps
            cap_functions = [max, min]
            correlation_factors = [0.99, 1.01]
            if not info.correlated:
                pass
            else:  # What is this doing? And why is it doing it? It will flip every time - which seems weird?
                correlation_factors.reverse()
                cap_functions.reverse()
                caps.reverse()

            # Rattle update amplitude
            if hasattr(generator, attribute):
                amplitude = getattr(generator, attribute)
                self.sampler.tracker.rattle_amplitudes[walker.index] = amplitude
                if acceptance < 0.5:
                    setattr(generator, attribute, cap_functions[0](caps[0], amplitude * correlation_factors[0]))
                else:
                    setattr(generator, attribute, cap_functions[1](caps[1], amplitude * correlation_factors[1]))
            
            self.writer.debug(f'{walker.index}: {attribute} = {getattr(generator, attribute)}')



from copy import deepcopy as dcpy
from typing import Dict, Optional

import numpy as np
from ase import Atoms

from agox.candidates import StandardCandidate
from agox.utils.thermodynamics import ThermodynamicsData


def get_composition(numbers: np.array, references: dict) -> dict:
    ##Gets sorted array of numbers containing species or molecules that contribute to gfe, longer first
    size_sort = np.argsort([-len(Atoms(k).numbers) for k in references.keys()])
    r_nums = [Atoms(k).numbers for k in references.keys()]
    r_keys = [k for k in references.keys()]
    nums = dcpy(numbers)
    ret = dict()
    for sz in size_sort:
        unique, counts = np.unique(nums, return_counts=True)
        d = dict(zip(unique, counts))
        hits = min([d.get(i, 0) for i in r_nums[sz]])
        ret[r_keys[sz]] = hits
        # delete from array
        for i in r_nums[sz]:
            idx = np.flatnonzero(nums == i)
            np.delete(nums, idx[:hits])
    return ret


def gibbs_free_energy(
    candidate: Optional[StandardCandidate] = None,
    numbers: Optional[np.array] = None,
    total_energy: float = None,
    template_energy: float = None,
    thermo_data: Optional[ThermodynamicsData] = None,
    references: Optional[Dict] = None,
    chemical_potentials: Optional[Dict] = None,
) -> float:
    """
    Calculate the Gibbs free energy of a candidate. 

    Parameters
    -----------
    candidate: StandardCandidate
        The candidate to calculate the Gibbs free energy for. Can be omitted 
        if numbers and total_energy are provided.
    numbers: np.array
        Array of the atomic numbers of the candidate. Can be omitted if candidate
        is provided.
    total_energy: float
        The total energy of the candidate. Can be omitted if candidate is provided.
    thermo_data: ThermodynamicsData
        The thermodynamic data to use for the calculation. Can be omitted if 
        references and chemical_potentials are provided.
    references: dict
        The references to use for the calculation. Can be omitted if thermo_data
        is provided.
    chemical_potentials: dict
        The chemical potentials to use for the calculation. Can be omitted if 
        thermo_data is provided.
    """

    if total_energy is None:
        total_energy = candidate.get_total_energy()
    if numbers is None:
        numbers = candidate.get_search_numbers()

    if thermo_data is not None:
        if references is None:
            references = thermo_data.references
        if chemical_potentials is None:
            chemical_potentials = thermo_data.chemical_potentials
        if template_energy is None:
            template_energy = thermo_data.template_energy
    else:
        if references is None or chemical_potentials is None or template_energy is None:
            raise ValueError("Thermodynamic data or references, chemical_potentials and template_energy must be provided")

    gibbs_energy = total_energy - template_energy
    comp = get_composition(numbers, references)

    for k in comp.keys():
        n = comp.get(k, 0)
        gibbs_energy -= n * (references[k] + chemical_potentials[k])

    return gibbs_energy



"""
Many algorithms use generators that are seeded by previous candidates, e.g. in
Basin-Hopping the next candidate is generated by a small random perturbation of
the previous candidate.

Samplers are responsible for keeping track of which candidates to use as seeds
for the next generation of candidates.
"""
# ruff: noqa: I001, E402
from typing import TypeAlias
from .ABC_sampler import SamplerBaseClass
Sampler: TypeAlias = SamplerBaseClass

from .concurrent_tempering import ConcurrentTemperingSampler
from .fixed_sampler import FixedSampler
from .genetic import DistanceComparator, GeneticSampler
from .kernel_similarity import KernelSimSampler
from .kmeans import KMeansSampler
from .metropolis import MetropolisSampler
from .replica_exchange import ReplicaExchangeSampler
from .spectral_graph import SpectralGraphSampler
from .gfe_kmeans import GFEKMeansSampler
from .gfe_sampler import GibbsFreeEnergySampler

__all__ = ['SamplerBaseClass',
            'SpectralGraphSampler',
            'KMeansSampler',
            'MetropolisSampler',
            'ConcurrentTemperingSampler',
            'GeneticSampler',
            'DistanceComparator',
            'KernelSimSampler',
            'FixedSampler',
            'GibbsFreeEnergySampler',
            'GFEKMeansSampler',
            'ReplicaExchangeSampler'
        ]


from typing import Callable, List

import numpy as np
import ray


@ray.remote(
    num_cpus=1,
    runtime_env={
        "env_vars": {
            "OMP_NUM_THREADS": "1",
            "OPENBLAS_NUM_THREADS": "1",
            "MKL_NUM_THREADS": "1",
        }
    },
    max_restarts=5,
)
class Actor:
    """
    Basic Ray Actor used by AGOX.
    """

    def __init__(self, actor_id: int = 0):
        self.modules = {}
        self.id = actor_id

    def execute_function(self, fn: Callable, module_keys: List, *args, **kwargs):
        """
        Execute a function on the actor.

        Parameters:
        -----------
        fn: function
            Function to execute.
        module_keys: list
            List of keys to modules to pass to the function.
        args: list
            Arguments to pass to the function.
        kwargs: dict
            Keyword arguments to pass to the function.
        """
        kwargs = self.set_seed(**kwargs)
        return fn(*[self.modules[key] for key in module_keys], *args, **kwargs)

    def set_seed(self, **kwargs):
        """
        Set the seed for the actor.
        """
        seed = kwargs.pop("pool_internal_seed", None)
        if seed is not None:
            np.random.seed(seed)
        return kwargs

    def add_module(self, module, key):
        """
        Add a module to the actor.

        Parameters:
        -----------
        module: AGOX module
            The module to add to the actor.
        key: module.ray_key (uuid).
            The key that references the module on the actors.
        """
        self.modules[key] = module

    def remove_module(self, key):
        """
        Remove a module from the actor.
        """
        del self.modules[key]



from collections import OrderedDict
from typing import Dict

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import rich_click as click
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from rich import print

from agox.analysis.search_analysis import SearchAnalysis


class EventHandler:
    def __init__(self, search_analysis: SearchAnalysis, axis_dict: Dict[str, Axes], fig: Figure) -> None:
        self.search_analysis = search_analysis
        self.axis_dict = axis_dict
        self.fig = fig

        self.total_images = search_analysis.get_number_of_best_candidates()
        self.image = 0
        self.plane = "xy"
        self.timer = self.fig.canvas.new_timer(interval=100)
        self.timer.add_callback(self.image_update_arrow_key)


    def on_press(self, event: matplotlib.backend_bases.KeyEvent) -> None:
        if event.key == "z":
            self.plane = "xy"
        elif event.key == "x":
            self.plane = "yz"
        elif event.key == "y":
            self.plane = "xz"
        elif event.key == "right":
            self.image += 1
        elif event.key == "left":
            self.image -= 1

        self.image = self.image % self.total_images

        if event.key in ["right", "left", "shift+right", "shift+left"]:
            if "right" in event.key:
                self.direction = 1
            else:
                self.direction = -1
            if "shift" in event.key:
                self.timer.start()
        else:
            self.image_update()

    def on_release(self, event: matplotlib.backend_bases.KeyEvent) -> None:
        if event.key in ["shift+right", "shift+left"]:
            self.timer.stop()
        self.image_update()

    def image_update_arrow_key(self) -> None:
        self.image += self.direction
        self.image = self.image % self.total_images
        self.image_update()

    def image_update(self) -> None:
        self.axis_dict["configuration"].clear()
        self.search_analysis.plot_configuration(ax=self.axis_dict["configuration"], image=self.image, plane=self.plane, 
                                                name= f"Candidate {self.image + 1}/{self.total_images}")

        if "distribution" in self.axis_dict:
            self.search_analysis.update_hist_line(image=self.image)

        self.fig.canvas.draw()


option_groups = {"agox analysis": []}
option_groups["agox analysis"].append({"name": "File options", "options": ["--reload"]})
option_groups["agox analysis"].append(
    {
        "name": "Plot options",
        "options": [
            "--backend",
            "--success",
            "--configuration",
            "--energy",
            "--distribution",
        ],
    }
)
option_groups["agox analysis"].append(
    {
        "name": "Analysis options",
        "options": ["--delta-total", "--delta-atom", "--property-name", "--thermodynamics-file", "--criterion"],
    }
)
click.rich_click.OPTION_GROUPS.update(option_groups)


@click.command("analysis")  # fmt: skip
@click.argument("directories", nargs=-1, type=click.Path(exists=True))  # fmt: skip
@click.option("--reload", "-r", is_flag=True, help="Reload the searches", default=False, show_default=True)  # fmt: skip
@click.option("--backend", "-b", type=str, help="Matplotlib backend", default="TkAgg", show_default=True)  # fmt: skip
@click.option("--success", "-s",help="Plot CDF (success curve).", is_flag=True, default=True, show_default=True, type=bool)  # fmt: skip
@click.option("--configuration", "-c", help="Plot the structures.", is_flag=True, default=True, show_default=True, type=bool)  # fmt: skip
@click.option("--energy", "-e", help="Plot the energies.", is_flag=True, default=True, show_default=True, type=bool)  # fmt: skip
@click.option("--distribution", "-d",help="Plot the distribution.", is_flag=True, default=False, show_default=True, type=bool)  # fmt: skip
@click.option("--delta-total", "-dE", help="Threshold for success in eV. Used if criteria is 'energy'.", type=float, default=1.0, show_default=True)  # fmt: skip
@click.option("--delta-atom", "-de", help="Threshold for success in eV/atom, overwrites delta-total if given", type=float, default=np.nan, show_default=True)  # fmt: skip
@click.option("--property-name", "-p", type=click.Choice(["Energy", "FreeEnergy"], case_sensitive=False), default="Energy", show_default=True, help="Property to consider")  # fmt: skip
@click.option("--criterion", "-C", type=click.Choice(["energy", "graph"], case_sensitive=False), default="energy", show_default=True, help="Criterion used to calculate success curve.")  # fmt: skip
@click.option("--thermodynamics-file", "-t", type=click.Path(exists=True), help="The thermodynamics json file that specifies references and chemical potentials to use for free energy", default=None)  # fmt: skip
@click.option("--window", "-w", type=int, help="Window for rolling average", default=5, show_default=True)  # fmt: skip
@click.option("--iteration-time", "-it", is_flag=True, help="Plot as the time axis being iterations", default=False, show_default=True)  # fmt: skip
@click.option("--time-time", '-tt', is_flag=True, help="Plot as the time axis being time", default=False, show_default=True)  # fmt: skip
@click.option('--time-unit', '-tu', type=click.Choice(['s', 'm', 'h', 'd', 'y'], case_sensitive=False), default='s', show_default=True, help='Time axis unit')  # fmt: skip
@click.option('--n_cores', '-nc', type=int, default=1, show_default=True, help='Number of cores to account for')  # fmt: skip
def cli_analysis(
    directories: str,
    reload: bool,
    backend: str,
    success: bool,
    configuration: bool,
    energy: bool,
    distribution: bool,
    delta_total: float,
    delta_atom: float,
    property_name: str,
    criterion: str,
    thermodynamics_file: str,
    window: int,
    iteration_time: bool,
    time_time: bool,
    time_unit: str,
    n_cores: int,
) -> None:
    """
    Perform basic analysis of searches.
    """
    from timeit import default_timer as dt

    from agox.analysis import DistanceCriterion, SearchCollection, ThresholdCriterion
    from agox.analysis.property import DescriptorProperty, EnergyProperty, FreeEnergyProperty
    from agox.analysis.search_analysis import SearchAnalysis
    from agox.models.descriptors import SpectralGraphDescriptor

    t0 = dt()
    matplotlib.use(backend)

    # Setup the search collection
    search_collection = SearchCollection(directories, reload=reload, progress_bar=True)
    print("Time taken to load: ", dt() - t0)

    for search in search_collection:
        print(search)

    # Check if we have anything to plot:
    plot_labels = ["configuration", "success", "energy", "distribution"]
    plot_bools = [configuration, success, energy, distribution]
    plotting_dict = OrderedDict()
    for label, plot_bool in zip(plot_labels, plot_bools):
        if plot_bool:
            plotting_dict[label] = plot_bool

    if len(plotting_dict) == 0:
        print("Nothing to plot.")
        return

    # Setup the figure:
    n_plots = len(plotting_dict)
    fig, axes = plt.subplots(1, n_plots, figsize=(4 * n_plots, 4), layout="constrained")
    axes = np.atleast_1d(axes)
    axes_dict = {}

    for ax, (label, plot_bool) in zip(axes, plotting_dict.items()):
        if plot_bool:
            axes_dict[label] = ax

    # Property:
    if iteration_time:
        time_axis = 'iterations'
    elif time_time:
        time_axis = 'time'
    else:
        time_axis = 'indices'

    if thermodynamics_file is not None:
        property_name = "FreeEnergy"  # If thermodynamics file is given, we assume we want to plot free energy.

    if property_name == "Energy":
        energy_property = EnergyProperty(time_axis=time_axis, time_unit=time_unit, cost_factor=n_cores)
        delta = delta_total if np.isnan(delta_atom) else delta_atom * len(search_collection[0].get_candidate(0, 0))
    elif property_name == "FreeEnergy":
        from agox.utils.thermodynamics import ThermodynamicsData
        thermo_data = ThermodynamicsData.load(thermodynamics_file)
        energy_property = FreeEnergyProperty(thermo_data=thermo_data, time_axis=time_axis)
        delta = delta_total

    search_analysis = SearchAnalysis(search_collection, energy_property)

    if configuration:
        search_analysis.plot_configuration(ax=axes_dict["configuration"], image=0)

    if energy:
        search_analysis.plot_energy(ax=axes_dict["energy"], window=window)

    if success:
        if criterion == "energy":
            _, _, min_props = search_analysis.get_flat_candidates()
            e_min = np.nanmin(min_props)
            criterion = ThresholdCriterion(e_min + delta)
            search_property = energy_property
        elif criterion == "graph":
            best_candidate = search_analysis.get_flat_candidates()[0][0]
            descriptor = SpectralGraphDescriptor.from_atoms(best_candidate)
            comparate = descriptor.get_features(best_candidate)
            search_property = DescriptorProperty(descriptor)
            criterion = DistanceCriterion(threshold=1e-7, comparate=comparate)  # Graphs should match.

        search_analysis.plot_cdf(ax=axes_dict["success"], criterion=criterion, search_property=search_property)

    if distribution:
        search_analysis.plot_histogram(ax=axes_dict["distribution"], bin_size=delta, image=0)

    # Setup the event handler:
    event_handler = EventHandler(search_analysis, axes_dict, fig)
    fig.canvas.mpl_connect("key_press_event", event_handler.on_press)
    fig.canvas.mpl_connect("key_release_event", event_handler.on_release)

    plt.show()



from typing import Dict, List

import numpy as np
import pytest
from ase import Atoms
from ase.calculators.singlepoint import SinglePointCalculator

from agox.candidates import StandardCandidate
from agox.utils.thermodynamics import ThermodynamicsData


@pytest.fixture(scope='module', params=[1, 2, 3])
def size(request: pytest.FixtureRequest) -> int:
    return request.param

@pytest.fixture(scope='module')
def total_energy(size: int) -> float:
    return -10.0 * size

@pytest.fixture(scope='module')
def formula(size: int) -> str:
    return f"Cu{size}O{2*size}"

@pytest.fixture(scope='module')
def symbols(size: int) -> List[str]:
    return ['Cu'] * size + ['O'] * 2*size

@pytest.fixture(scope='module')
def numbers(size: int) -> np.array:
    return np.array([29] * size + [8] * 2*size)

@pytest.fixture(scope='module')
def candidate(size: int, formula: str, total_energy: float) -> StandardCandidate:
    template = Atoms('', cell=np.eye(3)*10)
    atoms = Atoms(formula, positions=np.random.rand(size*3, 3)*10, cell=np.eye(3)*10)
    calculator = SinglePointCalculator(atoms, energy=total_energy)
    atoms.calc = calculator
    return StandardCandidate.from_atoms(template, atoms)

@pytest.fixture(scope='module')
def references() -> Dict[str, float]:
    return {'Cu': -1.0, 'O': -2.0}

@pytest.fixture(scope='module')
def chemical_potentials() -> Dict[str, float]:
    return {'Cu': -1.0, 'O': -2.0}

@pytest.fixture(scope='module')
def template_energy() -> float:
    return -10.0

@pytest.fixture(scope='module')
def thermo_data(references: Dict[str, float], chemical_potentials: Dict[str, float], template_energy: float) -> ThermodynamicsData:
    return ThermodynamicsData(references, chemical_potentials, template_energy)

@pytest.fixture(scope='module')
def expected_energy(symbols: List[str], total_energy: float, references: Dict[str, float], chemical_potentials: Dict[str, float], template_energy: float) -> float:

    gibbs = total_energy
    gibbs -= template_energy

    for symbol in symbols:
        gibbs -= references[symbol] + chemical_potentials[symbol]
    return gibbs



from dataclasses import dataclass

import numpy as np
from ase.calculators.singlepoint import SinglePointCalculator

from agox.candidates import Candidate


@dataclass
class SampleMember:
    _candidate: Candidate = None
    _energy: float = np.inf

    def empty(self) -> bool:
        return self.candidate is None
    
    def get_energy(self) -> float:
        return self._energy
    
    @property
    def candidate(self) -> Candidate:
        if self._candidate is None:
            return None
        candidate = self._candidate.copy()
        candidate.calc = SinglePointCalculator(candidate, **self._candidate.calc.results)
        return candidate
    
    @candidate.setter
    def candidate(self, value: Candidate) -> None:
        if isinstance(value, Candidate) or value is None:
            self._candidate = value
        else:
            raise ValueError('Candidate must be an instance of Candidate')

    @property
    def energy(self) -> float:
        return self._energy
    
    @energy.setter
    def energy(self, value: float) -> None:
        if isinstance(value, (int, float)) or value is None:
            self._energy = value
        else:
            raise ValueError('Energy must be a float or int')

    def set_index(self, i): 
        self._candidate.meta_information["walker_index"] = i


class ReplicaExchangeSample:

    def __init__(self, sample_size: int) -> None:
        self.sample_size = sample_size
        self.sample = [SampleMember() for _ in range(sample_size)]

    def swap(self, i: int, j: int) -> None:
        self.sample[i], self.sample[j] = self.sample[j], self.sample[i]
        self.sample[i].set_index(i)
        self.sample[j].set_index(j)
        
    def update(self, key: int, value: Candidate, energy: float) -> None:   
        if value is None: 
            _value = None
            energy = np.inf
        else:
            _value = value.copy()
            _value.calc = SinglePointCalculator(_value, **value.calc.results)
        self.sample[key].candidate = _value
        self.sample[key].energy = energy

    def __getitem__(self, key: int) -> Candidate:
        return self.sample[key]
    
    def __setitem__(self, key: int, value: Candidate) -> None:
        self.update(key, value, value.get_potential_energy())

    def __len__(self) -> int:
        return len(self.sample)
    
    def __iter__(self) -> iter:
        return iter(self.sample)


from ase.calculators.singlepoint import SinglePointCalculator as SPC

from agox.candidates.ABC_candidate import CandidateBaseClass


class StandardCandidate(CandidateBaseClass):
    """
    Standard candidate class.
    """

    @classmethod
    def from_atoms(cls, template, atoms):
        candidate = cls(
            template=template, positions=atoms.positions, numbers=atoms.numbers, cell=atoms.cell, pbc=atoms.pbc
        )
        if hasattr(atoms, "calc"):
            if atoms.calc is not None:
                if "energy" in atoms.calc.results:
                    if "forces" in atoms.calc.results:
                        candidate.calc = SPC(
                            candidate,
                            energy=atoms.get_potential_energy(apply_constraint=False),
                            forces=atoms.get_forces(apply_constraint=False),
                        )
                    else:
                        candidate.calc = SPC(candidate, energy=atoms.get_potential_energy(apply_constraint=False))
        return candidate



import functools
from typing import List

# from agox import __version__

ICON = """
       _            _  _  _        _  _  _  _    _           _ 
     _(_)_       _ (_)(_)(_) _   _(_)(_)(_)(_)_ (_)_       _(_)
   _(_) (_)_    (_)         (_) (_)          (_)  (_)_   _(_)  
 _(_)     (_)_  (_)    _  _  _  (_)          (_)    (_)_(_)    
(_) _  _  _ (_) (_)   (_)(_)(_) (_)          (_)     _(_)_     
(_)(_)(_)(_)(_) (_)         (_) (_)          (_)   _(_) (_)_   
(_)         (_) (_) _  _  _ (_) (_)_  _  _  _(_) _(_)     (_)_ 
(_)         (_)    (_)(_)(_)(_)   (_)(_)(_)(_)  (_)         (_)  v{}_{} \n
"""


def get_git_revision_short_hash() -> str:
    import os
    import subprocess

    import agox

    try:
        dir_path = os.path.dirname(agox.__path__[0])
        version_string = (
            subprocess.check_output(["git", f"--git-dir={dir_path}/.git", "rev-parse", "--short", "HEAD"])
            .decode("ascii")
            .strip()
        )
    except subprocess.CalledProcessError:
        version_string = "unknown"
    except BlockingIOError:
        version_string = "unknown"
    except FileNotFoundError:
        version_string = "unknown"

    return version_string


LINE_LENGTH = 79
PADDING_CHARACTER = "="
TERMINATE_CHARACTER = "|"


def get_icon():
    # version_string = get_git_revision_short_hash()
    return ICON


from abc import ABC, abstractmethod

from agox.observer import Observer
from agox.writer import Writer


class PostprocessBaseClass(ABC, Observer):
    """
    Base class for all postprocessors.

    Postprocessors are used to apply common postprocessing steps to all generated candidates.

    Parameters
    ----------
    gets : dict
        Dictionary of get keys, e.g. {'get_key':'candidates'}. Used to select
        from which entry in the agox.main.State cache the postprocessor should get candidates.
    sets : dict
        Dictionary of set keys, e.g. {'set_key':'candidates'}. Used to select
        in which entry in the agox.main.State cache the postprocessor should store the postprocessed candidates.
    """

    def __init__(
        self,
        gets: dict = None,
        sets: dict = None,
        order: int = 3,
        surname: str = None,
        **kwargs,
    ) -> None:
        if gets is None:
            gets = {"get_key": "candidates"}
        if sets is None:
            sets = {"set_key": "candidates"}        

        Observer.__init__(self, gets=gets, sets=sets, order=order, surname=surname, **kwargs)

        self.add_observer_method(
            self.postprocess_candidates,
            sets=self.sets[0],
            gets=self.gets[0],
            order=self.order[0],
            handler_identifier="AGOX",
        )

    def update(self):
        """
        Used if the postprocessor needs to continously update, e.g. the training of a surrogate potential.
        """
        pass

    @abstractmethod
    def postprocess(self, candidate):  # pragma: no cover
        """
        Method that actually do the post_processing
        """
        return postprocessed_candidate

    def process_list(self, list_of_candidates):
        """
        This allows all postproccesors to act on a list of candidates serially.
        This function can be overwritten by sub-class to implement parallelism.
        """
        processed_candidates = []
        for candidate in list_of_candidates:
            processed_candidate = self.postprocess(candidate)
            processed_candidates.append(processed_candidate)
        return processed_candidates

    def __add__(self, other):
        return SequencePostprocess(processes=[self, other], order=self.order)

    @Observer.observer_method
    def postprocess_candidates(self, state):
        candidates = state.get_from_cache(self, self.get_key)

        if self.do_check():
            candidates = self.process_list(candidates)
            candidates = list(filter(None, candidates))

        # Add data in write mode - so overwrites!
        state.add_to_cache(self, self.set_key, candidates, mode="w")


class SequencePostprocess(PostprocessBaseClass):
    name = "PostprocessSequence"

    def __init__(self, processes=[], order=None):
        self.processes = processes
        self.order = order

    def postprocess(self, candidate):
        for process in self.processes:
            candidate = process.postprocess(candidate)

        return candidate

    def process_list(self, list_of_candidates):
        for process in self.processes:
            list_of_candidates = process.process_list(list_of_candidates)
        return list_of_candidates

    def __add__(self, other):
        self.processes.append(other)
        return self

    def attach(self, main):
        for j, process in enumerate(self.processes):
            process.update_order(process.postprocess_candidates, order=self.order[0] + j * 0.1)
            process.attach(main)



from typing import Dict, List, Optional, Union

import numpy as np
from ase import Atoms
from matplotlib.axes import Axes

from agox.analysis.criterion import BaseCriterion
from agox.analysis.plot import PropertyPlotter, SuccessPlotter
from agox.analysis.property import Property
from agox.analysis.search_data import SearchCollection
from agox.candidates import StandardCandidate
from agox.utils.plot import plot_atoms, plot_cell
from agox.utils.plot.plot_atoms import plane_to_indices


def align(atoms: Union[Atoms, StandardCandidate]) -> Atoms:
    from sklearn.decomposition import PCA

    pca = PCA(n_components=3)
    pca.fit(atoms.get_positions())
    atoms.positions = pca.transform(atoms.get_positions())
    atoms.center()
    return atoms


class SearchAnalysis:
    def __init__(
        self, search_collection: Optional[SearchCollection] = None, analysis_property: Optional[Property] = None
    ) -> None:
        """
        Helper class for analyzing search results. Mainly used in the CLI.
        Functionality should be placed in different modules, this just collects
        them and joins them together for the CLI.

        Parameters:
        -----------
        search_collection: SearchCollection
            The search collection to analyze.
        analysis_property: Property
            The property to analyze.
        """

        self.search_collection = search_collection
        self.analysis_property = analysis_property

    def get_property(self, recalculate: bool = False) -> List:
        if not hasattr(self, "properties") or recalculate:
            properties = []
            for search in self.search_collection:
                # Gather the properties for all searches:
                properties.append(self.analysis_property(search))
            self.properties = properties

        return self.properties

    def get_identifiers(self, recalculate: bool = False) -> Dict:
        if not hasattr(self, "identifiers") or recalculate:
            identifiers = {}
            for search in self.search_collection:
                restart_identifiers = search.get_all_identifiers()
                search_label = search.get_label()
                search_labels = [f"{search_label}/{identifier}" for identifier in restart_identifiers]
                identifiers[search.get_uuid()] = search_labels

            self.identifiers = identifiers
        return self.identifiers

    def get_best_indices(self, recalculate: bool = False) -> Dict[str, int]:
        if not hasattr(self, "best_indices") or recalculate:
            properties = self.get_property(recalculate=recalculate)
            property_values = [prop.data for prop in properties]
            best_indices = {search.get_uuid(): [] for search in self.search_collection}
            for prop_array, search in zip(property_values, self.search_collection):
                for restart in range(search.get_number_of_restarts()):
                    min_index = np.nanargmin(prop_array[restart])
                    best_indices[search.get_uuid()].append(min_index)
            self.best_indices = best_indices
        return self.best_indices

    def get_best_property(self, recalculate: bool = False) -> Dict:
        indices = self.get_best_indices(recalculate=recalculate)
        properties = self.get_property(recalculate=recalculate)
        property_arrays = [prop.data for prop in properties]

        best_properties = {search.get_uuid(): [] for search in self.search_collection}
        for prop_array, search in zip(property_arrays, self.search_collection):
            for restart, min_index in enumerate(indices[search.get_uuid()]):
                best_properties[search.get_uuid()].append(prop_array[restart][min_index])
        return best_properties

    def get_candidates(self, recalculate: bool = False) -> Dict:
        if not hasattr(self, "candidates") or recalculate:
            candidates = []
            for search in self.search_collection:
                for restart, index in self.get_best_indices()[search.get_uuid()]:
                    candidates.append(search.get_candidate(restart, index))
            self.candidates = candidates
        return self.candidates

    def get_best_candidates(self, recalculate: bool = False) -> Dict:
        if not hasattr(self, "best_candidates") or recalculate:
            best_candidates = {search.get_uuid(): [] for search in self.search_collection}
            indices = self.get_best_indices()
            for search in self.search_collection:
                for restart, index in enumerate(indices[search.get_uuid()]):
                    best_candidates[search.get_uuid()].append(search.get_candidate(restart, index))
            self.best_candidates = best_candidates

        return self.best_candidates

    def get_number_of_best_candidates(self) -> int:
        return sum([len(candidates) for candidates in self.get_best_candidates().values()])

    def get_flat_candidates(self, recalculate: bool = False) -> List:
        if not hasattr(self, "flat_candidates") or recalculate:
            identifiers = self.get_identifiers()
            properties = self.get_best_property()
            candidates = self.get_best_candidates()

            flat_properties = []
            flat_identifiers = []
            flat_candidates = []

            for label, prop in properties.items():
                flat_properties.extend(prop)
                flat_identifiers.extend(identifiers[label])
                flat_candidates.extend(candidates[label])

            indices = np.argsort(flat_properties)
            flat_properties = [flat_properties[i] for i in indices]
            flat_identifiers = [flat_identifiers[i] for i in indices]
            flat_candidates = [flat_candidates[i] for i in indices]

            self.flat_candidates = flat_candidates
            self.flat_identifiers = flat_identifiers
            self.flat_properties = flat_properties
        return self.flat_candidates, self.flat_identifiers, self.flat_properties

    def plot_configuration(
        self,
        ax: Axes,
        candidate: Union[Atoms, StandardCandidate] = None,
        image: Optional[int] = None,
        plane: str = "xy",
        name: Optional[str] = None,
    ) -> None:
        """
        Plot the candidates of the search.

        Parameters:
        -----------
        ax: Axes
            The axis to plot on.
        candidate: Atoms
            The candidate to plot.
        image: int
            The index of the candidate to plot.
        plane: str
            The plane to plot.
        """

        name = name or ""

        if candidate is None:
            flat_candidates, flat_identifiers, flat_properties = self.get_flat_candidates()
            candidate = flat_candidates[image]
            if not candidate.pbc.any():
                candidate = align(candidate)
            name = f"{name} {flat_identifiers[image]}"
            energy = flat_properties[image]
        else:
            energy = candidate.get_potential_energy()

        plot_atoms(ax, candidate, repeat=int(candidate.pbc.any()), plane=plane)
        plot_cell(ax, candidate.get_cell(), plane=plane)
        ax.set_title(f"{name} \n Energy: {energy:.2f} eV")

        # Plot coordinate system:
        ax.arrow(0.1, 0.1, 0.1, 0, transform=ax.transAxes, color='black', linewidth=1.5, head_width=0.01, head_length=0.01)
        ax.arrow(0.1, 0.1, 0.0, 0.1, transform=ax.transAxes, color='black', linewidth=1.5, head_width=0.01, head_length=0.01)
        a, b, c, d = plane_to_indices(plane)
        indices_to_text = {0: 'x', 1: 'y', 2: 'z'}
        ax.text(0.22, 0.1, indices_to_text[a], transform=ax.transAxes, color='black', fontsize=10, verticalalignment='center', horizontalalignment='left')
        ax.text(0.1, 0.22, indices_to_text[b], transform=ax.transAxes, color='black', fontsize=10, verticalalignment='bottom', horizontalalignment='center')


    def plot_energy(self, ax: Axes, window=5) -> None:
        """
        Plot the energy versus time.
        """
        # This recalculates - so thats annoying atm.
        property_time = PropertyPlotter(self.search_collection, self.analysis_property, window=window)
        property_time.plot(ax)

    def plot_cdf(self, ax: Axes, criterion: BaseCriterion, search_property: Optional[Property] = None) -> None:
        """
        Plot the CDF of the energy.
        """
        if search_property is None:
            search_property = self.analysis_property

        plotter = SuccessPlotter(self.search_collection, search_property, criterion)
        plotter.plot(ax)

    def plot_histogram(self, ax: Axes, bin_size: float, image: int = 0) -> None:
        best_properties = self.get_best_property()

        _, _, flat_best_properties = self.get_flat_candidates()
        bins = np.arange(np.min(flat_best_properties), np.max(flat_best_properties) + bin_size, bin_size)

        for label, values in best_properties.items():
            ax.hist(values, label=label, bins=bins, alpha=0.5, edgecolor="black")

        self.hist_line = ax.axvline(flat_best_properties[image], label="_no_legend_", color="black")

        ax.legend()
        ax.set_title("Histogram of energies")

    def update_hist_line(self, image: int = 0) -> None:
        line = self.hist_line
        _, _, flat_best_properties = self.get_flat_candidates()
        line.set_xdata([flat_best_properties[image]])



"""
A database records the candidates that have been evaluated during a search.

Further, because some things need to be trained, e.g. a ML potential model,
databases can be used to store the training data. The databases all implement
a similar observer-pattern as the main AGOX class, so models can be trained
whenever data is added to the database.
"""
from .ABC_database import DatabaseBaseClass
from agox.databases.database import Database
from agox.databases.database_concurrent import ConcurrentDatabase


__all__ = ["DatabaseBaseClass", "Database", "ConcurrentDatabase"]



from .observer_method import ObserverMethod  # noqa
from .observer_handler import ObserverHandler
from .finalization_handler import FinalizationHandler
from .observer import Observer

__all__ = ["ObserverHandler", "FinalizationHandler", "Observer", "ObserverMethod"]




from agox.acquisitors.LCB import LowerConfidenceBoundAcquisitor


class PowerLowerConfidenceBoundAcquisitor(LowerConfidenceBoundAcquisitor):
    name = "PowerLowerConfindenceBoundAcquisitor"

    def __init__(self, power=0, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.power = power

    def acquisition_function(self, E, sigma):
        return E - self.kappa * sigma**self.power

    def acquisition_force(self, E, F, sigma, sigma_force):
        return F - self.kappa * self.power * sigma ** (self.power - 1) * sigma_force



from typing import Callable

import numpy as np
from ase.calculators.calculator import all_properties
from ase.calculators.singlepoint import SinglePointCalculator
from ase.constraints import FixAtoms
from ase.optimize import BFGS
from ase.optimize.optimize import Optimizer

from agox.candidates import Candidate
from agox.models import Model
from agox.postprocessors.ABC_postprocess import PostprocessBaseClass
from agox.postprocessors.ray_relax.remote_relax import RelaxationResult, remote_relax
from agox.utils.ray import RayPoolUser, Task


class ParallelRelaxPostprocess(PostprocessBaseClass, RayPoolUser):
    """
    Relaxes candidates in parallel using Ray.

    Parameters
    ----------
    model : agox.models.ABC_model
        Model to use for relaxation.
    optimizer : ase.optimize.Optimizer
        Optimizer to use for relaxation.
    optimizer_run_kwargs : dict
        Keyword arguments to pass to the optimizer.run method.
    optimizer_kwargs : dict
        Keyword arguments to pass to the optimizer constructor.
    fix_template : bool
        If True, the template atoms are fixed during relaxation.
    constraints : list
        List of constraints to apply during relaxation.
    start_relax : int
        The iteration to start relaxing candidates.
    actor_message_function : callable
        Function that creates a message that will be printed. The function
        must takes: candidate, optimizer, model as arguments and return a string or
        a list of strings.
    trajectory: bool
        Whether or not trajectories are returned by the remote relaxation function.
    """

    name = "PoolRelaxer"

    def __init__(
        self,
        model: Model = None,
        optimizer: Optimizer | None = None,
        optimizer_kwargs: dict | None = None,
        optimizer_run_kwargs: dict | None = None,
        fix_template: bool = True,
        constraints: list | None = None,
        start_relax: int = 1,
        actor_message_function: Callable | None = None,
        trajectory: bool = False,
        trajectory_function: Callable | None = None,
        **kwargs,
    ) -> None:
        RayPoolUser.__init__(self)
        PostprocessBaseClass.__init__(self, **kwargs)

        # Optimizer settings:
        self.optimizer = BFGS if optimizer is None else optimizer
        self.optimizer_kwargs = optimizer_kwargs if optimizer_kwargs is not None else {"logfile": None}
        self.optimizer_run_kwargs = (
            optimizer_run_kwargs if optimizer_run_kwargs is not None else {"fmax": 0.05, "steps": 200}
        )
        self.start_relax = start_relax

        # Model
        self.model = model
        self.model_key = self.pool_add_module(model)

        # Constraints:
        self.constraints = constraints if constraints is not None else []
        self.fix_template = fix_template
        self.actor_message_function = actor_message_function

        # Dealing with trajectories.
        self.trajectory = trajectory
        self.set_trajectory_function(trajectory_function)

    def process_list(self, candidates: list[Candidate]) -> list[Candidate]:
        """
        Relaxes the given candidates in parallel using Ray.

        Parameters
        ----------
        list_of_candidates : listn
            List of AGOX candidates to relax.

        Returns
        -------
        list
            List of relaxed candidates.
        """

        if len(candidates) == 0:
            return candidates

        # Apply constraints.
        [self.apply_constraints(candidate) for candidate in candidates]

        # Make args, kwargs and modules lists:
        tasks = []
        for candidate in candidates:
            # Create task: function, modules, args, kwargs
            task = Task(
                function=remote_relax,
                modules=[self.model_key],  # The model is stored on the actors, tell the task where to find it.
                args=[candidate, self.optimizer, self.optimizer_kwargs, self.optimizer_run_kwargs],
                kwargs={"message_function": self.actor_message_function, "trajectory": self.trajectory},
            )
            tasks.append(task)

        relaxed_results = self.task_map(tasks)
        candidates = self._set_results(candidates, relaxed_results)
        self.report_basic_information(relaxed_results)
        self.report_messages(relaxed_results)

        if self.trajectory:
            self.apply_trajectory_function(relaxed_results)

        return candidates

    def postprocess(self, candidate: Candidate) -> Candidate:
        raise NotImplementedError('"postprocess"-method is not implemented, use postprocess_list.')

    def do_check(self, **kwargs) -> bool:
        return self.check_iteration_counter(self.start_relax)

    def _set_results(self, candidates: list[Candidate], results: list[RelaxationResult]) -> None:
        _candidates = []
        for candidate, result in zip(candidates, results):
            if result.relaxation_steps > 0:
                self.remove_constraints(candidate)
                self._set_result_to_candidate(candidate, result)
                _candidates.append(candidate)
        return _candidates

    def _set_result_to_candidate(self, candidate: Candidate, result: RelaxationResult) -> None:
        candidate.set_positions(result.candidate.positions)
        results = {prop: val for prop, val in result.candidate.calc.results.items() if prop in all_properties}
        candidate.calc = SinglePointCalculator(candidate, **results)
        candidate.meta_information.update(result.candidate.meta_information)

    def report_basic_information(self, results: list[RelaxationResult]) -> None:
        steps = np.array([result.relaxation_steps for result in results])
        convergence = np.array([result.converged for result in results])
        max_forces = np.array([result.max_force for result in results])
        improvement = np.array([result.final_energy - result.initial_energy for result in results])


        self.writer(f"{len(results)} candidates relaxed.")
        self.writer("Mean number of steps: {:.1f}".format(steps.mean()))
        self.writer("Std. number of steps: {:.1f}".format(steps.std()))
        self.writer("Mean \N{Greek Capital Letter Delta}E: {:.3f} eV".format(improvement.mean()))
        for idx, delta in enumerate(improvement):
            self.writer.debug(f"Relaxation {idx}: \N{Greek Capital Letter Delta}E = {delta}")
        self.writer(f"Mean max force: {max_forces.mean():.3f}")
        self.writer(f"Std. max force: {max_forces.std():.3f}")
        self.writer(f"Convergence: {convergence.sum()} / {len(convergence)}")

    def report_messages(self, results: list[RelaxationResult]) -> None:
        """
        Print messages from the relaxation results.

        Parameters
        ----------
        results : list
            List of relaxation results.
        """
        for idx, result in enumerate(results):
            if result.message is not None:
                message = "\n".join(result.message)
                self.writer(message)

    def apply_trajectory_function(self, results: list[RelaxationResult]) -> None:
        """
        Apply the trajectory function to the relaxation results. This method 
        calls the trajectory function with the trajectories and the iteration
        counter. May be used to e.g. save the trajectories to disk or print 
        some information about the trajectories.

        Parameters
        ----------
        results : list
            List of relaxation results.

        Returns
        -------
            None
        """
        trajectories = [result.trajectory for result in results]
        iteration = self.get_iteration_counter()
        self._trajectory_function(trajectories, iteration)

    @staticmethod
    def _default_trajectory_function(trajectories: list[list[Candidate]], iteration: int) -> None:
        """
        Default trajectory function that does nothing.
        """
        return None

    def set_trajectory_function(self, function: Callable) -> None:
        """
        Set the function that processes the relaxation trajectories. The function must take
        two arguments: trajectories, iteration. The function should not return anything.
        This will not be called unless `trajectory = True` in the constructor of this 
        class.
        
        Example: That stores the trajectories to disk.

        ```python
        def trajectory_function(trajectories: list[list[Candidate]], iteration: int):
            from ase.io import write
            for index, traj in enumerate(trajectories):
                if traj is not None:
                    write(f"trajectory_iter{iteration}_{index}.traj", traj)
        ```

        Parameters
        ----------
        function : callable
            Function that processes the relaxation trajectories.
        """
        self._trajectory_function = function if function is not None else self._default_trajectory_function


    def set_actor_message_function(self, actor_message_function: Callable) -> None:
        """
        Set the actor message function.

        Example:

        ```python
        def message_function(candidate, optimizer, model) -> str:
            import ray
            message = []
            message.append(f"{ray.get_runtime_context().get_actor_id()}")
            message.append(f"Max force: {candidate.get_forces().max()}")
            return message
        ```

        Parameters
        ----------
        actor_message_function : callable
            Function that creates a message that will be printed. The function
            must takes: candidate, optimizer, model as arguments and return a string or
            a list of strings.
        """
        self.actor_message_function = actor_message_function

    ####################################################################################################################
    # Constraints
    ####################################################################################################################

    def apply_constraints(self, candidate: Candidate) -> None:
        constraints = self.constraints + []
        if self.fix_template:
            constraints.append(self.get_template_constraint(candidate))

        for constraint in constraints:
            if hasattr(constraint, "reset"):
                constraint.reset()

        candidate.set_constraint(constraints)

    def remove_constraints(self, candidate: Candidate) -> None:
        candidate.set_constraint([])

    def get_template_constraint(self, candidate: Candidate) -> FixAtoms:
        return FixAtoms(indices=np.arange(len(candidate.template)))



import functools
from abc import ABC
from copy import deepcopy
from typing import Any, Callable, Tuple

import numpy as np
from ase import Atoms
from ase.calculators.singlepoint import SinglePointCalculator

from agox.module import Module
from agox.utils.cache import Cache


class CandidateBaseClass(ABC, Atoms, Module):
    """
    Base-class for Candidate objects.

    Parameters
    ------------
    template: ase.Atoms
        Atoms object of the template structure. Does not need to be supplied
        if 'template_indices' are given.
    template_indices: np.array
        Indices of template atoms.
    kwargs:
        Everything that can be supplied to an ASE atoms object, specifically
        cell, positions and numbers of ALL atoms - including template atoms.
    """

    def __init__(
        self, template: Atoms = None, template_indices: np.ndarray = None, use_cache: bool = True, **kwargs
    ) -> None:
        Atoms.__init__(self, **kwargs)  # This means all Atoms-related stuff gets set.
        Module.__init__(self, use_cache=use_cache)
        self.meta_information = self.info

        self.use_cache = use_cache
        self._cache = Cache()

        # Template stuff:
        if template_indices is not None:
            self.template_indices = template_indices.astype(int)
            self.template = self.get_template()
        elif template is not None:
            self.template = template
            self.template_indices = np.arange(len(template))
        else:
            print("You have not supplied a template, using an empty atoms object without PBC and no specified cell.")
            self.template = Atoms(pbc=self.pbc)
            self.template_indices = np.arange(0)

        self.set_pbc(self.template.get_pbc())  # Inherit PBC's from template.

    @classmethod
    def cache(cls, key: str) -> Callable:  # noqa: ANN001
        def decorator(func):  # noqa: ANN001
            @functools.wraps(func)
            def wrapper(self, atoms: Atoms, *args, **kwargs):  # noqa: ANN001
                if not self.use_cache:
                    return func(self, atoms, *args, **kwargs)

                full_key = self.cache_key + "/" + key
                if isinstance(atoms, CandidateBaseClass):
                    value = atoms.get_from_cache(full_key)
                    if value is None:
                        value = func(self, atoms, *args, **kwargs)
                        if atoms.use_cache:
                            atoms.set_to_cache(full_key, value)
                else:
                    value = func(self, atoms, *args, **kwargs)

                return value

            return wrapper

        return decorator

    def get_from_cache(self, key: str) -> Any:
        if not self.use_cache:
            return None

        identifier, value = self._cache.get(key, (None, None))
        if identifier is not None:
            if self.compare_identity(identifier):
                return value
        else:
            return None

    def set_to_cache(self, key: str, value: Any) -> None:
        self._cache.put(key, (self.get_identifier(), value))

    def compare_identity(self, identifier: Tuple) -> bool:
        for a, b in zip(identifier, self.get_identifier()):
            equal = (a == b).all()
            if not equal:
                return equal
        return equal

    def get_identifier(self) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        return (self.get_atomic_numbers(), self.get_positions(), self.get_cell())

    def add_meta_information(self, name: str, value: Any) -> None:
        """
        Adds an entry to the meta_information dictionary

        Parameters
        -----------
        name: str (preferably, but can take anything that indexes a dict)
            Key to the dict
        value:
            Value to be set in the dict.
        """
        self.meta_information[name] = value

    def add_write_key(self, value: Any) -> None:
        """
        Adds a write key to the candidates meta information.

        Parameters
        -----------
        value: str
            key to add to the write information

        """
        try:
            keys = self.get_meta_information("write_keys")
            if isinstance(keys, str):
                keys = list(keys.split(" "))
            if value not in keys:
                keys.append(value)
        except Exception:
            keys = [value]
        self.add_meta_information("write_keys", keys)

    def print_properties(self, writer: Callable, iteration: int) -> None:
        keys = self.get_meta_information("write_keys")
        strings_list = []
        write_string = "Energy: " + str(iteration) + "  " + str(self.get_potential_energy()) + "  "

        if keys is not None:
            for key in keys:
                addendum = "[" + key + ":" + " " + self.get_meta_information(key) + "] "
                if len(write_string) + len(addendum) > 120:
                    strings_list.append(write_string)
                    write_string = ""

                write_string += addendum
        strings_list.append(write_string)
        if strings_list is not None:
            for string in strings_list:
                writer(string)

    def get_meta_information(self, name: str) -> Any:
        """
        Get from the meta_information dict.

        Parameters
        -----------
        name: str
            Key to get with.

        Returns
        --------
        value - any type
            A copy of the wanted entry in the dict or None if it is not set.
        """
        try:
            return self.meta_information.get(name, None).copy()
        except AttributeError:
            # This catches for example 'int' that dont have a copy method.
            # Ints won't change in-place, but it perhaps custom classes will.
            return self.meta_information.get(name, None)

    def get_meta_information_no_copy(self, name: str) -> Any:
        """
        Get from the meta_information dict without copying

        Parameters
        -----------
        name: str
            Key to get with.

        Returns
        --------
        value - any type
            The wanted entry in the dict or None if it is not set.
        """
        return self.meta_information.get(name, None)

    def has_meta_information(self, name: str) -> bool:
        """
        Get from the meta_information dict without copying

        Parameters
        -----------
        name: str
            Key to get with.

        Returns
        --------
        bool
            True if the 'name' is a key to meta_information.
        """
        return name in self.meta_information.keys()

    def pop_meta_information(self, name: str) -> Any:
        """
        Pop from the meta_information dict.

        Parameters
        -----------
        name: str
            Key to get with.

        Returns
        --------
        value - any type
            The wanted entry in the dict or None if it is not set.
        """
        return self.meta_information.pop(name, None)

    def reset_meta_information(self) -> None:
        """
        Resets meta_information
        """
        self.meta_information = dict()

    def get_template(self) -> Atoms:
        """
        Get the template atoms object.

        Returns
        --------
        atoms
            Template as an Atoms object.
        """
        return Atoms(
            numbers=self.numbers[self.template_indices],
            positions=self.positions[self.template_indices],
            cell=self.cell,
            pbc=self.pbc,
        )

    def copy(self) -> "CandidateBaseClass":
        """
        Return a copy of candidate object.

        Not sure if template needs to be copied, but will do it to be safe.

        Returns
        --------
        candidate
            A copy of the candidate object.
        """
        candidate = self.__class__(
            template=self.template.copy(), cell=self.cell, pbc=self.pbc, info=self.info, celldisp=self._celldisp.copy()
        )
        candidate.meta_information = self.meta_information.copy()

        candidate.arrays = {}
        for name, a in self.arrays.items():
            candidate.arrays[name] = a.copy()

        candidate.constraints = deepcopy(self.constraints)

        return candidate

    def copy_calculator_to(self, atoms: Atoms) -> None:
        """
        Copy current calculator and attach to the atoms object
        """
        if self.calc is not None and "energy" in self.calc.results:
            if "forces" in self.calc.results:
                calc = SinglePointCalculator(
                    atoms, energy=self.calc.results["energy"], forces=self.calc.results["forces"]
                )
            else:
                calc = SinglePointCalculator(atoms, energy=self.calc.results["energy"])
            atoms.set_calculator(calc)

    def get_property(self, key: str) -> Any:
        """
        Get property from calculator.

        Parameters
        -----------
        key
            Key used to index calc.get_property

        Returns
        --------
        value
            Value of calc.get.property(key)
        """

        return self.calc.get_property(key)

    def get_template_indices(self) -> np.array:
        """
        Returns
        --------
        np.array
            Array of template indices
        """
        return self.template_indices

    def get_optimize_indices(self) -> np.array:
        """
        Returns
        --------
        np.array
            Indices of of atoms that are part of the search.
        """
        return np.arange(len(self.template), len(self))

    def get_center_of_geometry(self, all_atoms: bool = False) -> np.array:
        """
        Returns the center of geometry.

        Parameters
        -----------
        all_atoms: bool
            If True all atoms are included, if False only non-template atoms are.
        """

        if all_atoms:
            return np.mean(self.positions, axis=0).reshape(1, 3)
        else:
            return np.mean(self.positions[self.get_optimize_indices()], axis=0).reshape(1, 3)

    def set_calculator(self, calc: Any) -> None:
        """
        "Old" ASE syntax for setting calculator, avoids the annoying warning and
        uses this better syntax.

        Parameters
        -----------
        calc: calculator
            Calculator to set.
        """
        self.calc = calc

    def get_search_formula(self):
        """
        Get the formula of the search atoms.

        Returns
        --------
        str
            Chemical formula of the search atoms.
        """
        numbers = self.get_search_numbers()
        return Atoms(numbers=numbers).symbols.get_chemical_formula()

    def get_search_numbers(self):
        """
        Get the atomic numbers of the search atoms.

        Returns
        --------
        np.array
            Atomic numbers of the search atoms.
        """
        return self.get_atomic_numbers()[self.get_optimize_indices()]

    def __getitem__(self, i):
        """
        Get item from atoms object.

        Parameters
        -----------
        i: int, array-like or slice.
            Index to get.

        Returns
        --------
        atom
            Atom at index i.
        """
        import numbers

        if isinstance(i, numbers.Integral):
            return super().__getitem__(i)

        # All the indices that are to be extracted:
        if isinstance(i, slice):
            indices = np.arange(i.start, len(self), i.step)
        else:
            indices = np.array(i)

        # Determine if any are template atoms:
        template_indices = np.intersect1d(indices, self.template_indices)

        # If there are template atoms, we need to adjust the indices.
        template_indices = np.where(np.isin(indices, self.template_indices))[0]

        return self.__class__(
            template_indices=template_indices,
            positions=self.positions[indices].copy(),
            numbers=self.numbers[indices].copy(),
            cell=self.cell.copy(),
            pbc=self.pbc.copy(),
        )

    def set_constraint(self, constraints=None):
        """
        Set constraints to the candidate.

        Parameters
        -----------
        """
        from agox.utils.constraints import ConstraintManager

        if isinstance(constraints, ConstraintManager):
            constraints = constraints.get_constraints(self)
        super().set_constraint(constraints)


def disable_cache(candidate: CandidateBaseClass) -> None:
    # Can be either Candidate or Atoms.
    if isinstance(candidate, CandidateBaseClass):
        candidate.use_cache = False


def switch_cache(candidate: CandidateBaseClass, state: bool) -> bool:
    prev_state = False
    if isinstance(candidate, CandidateBaseClass):
        prev_state = candidate.use_cache
        candidate.use_cache = state
    return prev_state



from pathlib import Path
from typing import Dict, List, Union

import matplotlib
import matplotlib.backend_bases
import matplotlib.pyplot as plt
import rich_click as click
from ase import Atoms
from ase.io import read
from ase.visualize import view
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from rich import print

from agox.analysis.search_analysis import SearchAnalysis


class EventHandler:
    def __init__(
        self,
        search_analysis: SearchAnalysis,
        axis_dict: Dict[str, Axes],
        figure: Figure,
        configurations: List[Atoms],
        energy_plot: "EnergyPlot" = None,
    ) -> None:
        self.search_analysis = search_analysis
        self.axis_dict = axis_dict
        self.fig = figure

        self.energy_plot = energy_plot

        self.image = 0
        self.total_images = len(configurations)
        self.configurations = configurations
        self.plane = "xy"
        self.timer = self.fig.canvas.new_timer(interval=50)
        self.timer.add_callback(self.image_update_arrow_key)

    def on_press(self, event: matplotlib.backend_bases.KeyEvent) -> None:
        if event.key == "z":
            self.plane = "xy"
        elif event.key == "x":
            self.plane = "yz"
        elif event.key == "y":
            self.plane = "xz"
        elif event.key == "right":
            self.image += 1
        elif event.key == "left":
            self.image -= 1

        self.image = self.image % self.total_images

        if event.key in ["right", "left"]:
            if event.key == "right":
                self.direction = 1
            else:
                self.direction = -1
            self.timer.start()

    def on_release(self, event: matplotlib.backend_bases.KeyEvent) -> None:
        if event.key in ["right", "left"]:
            self.timer.stop()
        self.image_update()

    def image_update_arrow_key(self) -> None:
        self.image += self.direction
        self.image = self.image % self.total_images
        self.image_update()

    def image_update(self) -> None:
        self.axis_dict["configuration"].clear()
        self.search_analysis.plot_configuration(
            ax=self.axis_dict["configuration"],
            candidate=self.configurations[self.image],
            plane=self.plane,
            name=self.image,
        )

        if self.energy_plot is not None:
            self.energy_plot.update(self.image)

        self.fig.canvas.draw()


class EnergyPlot:
    def __init__(self, energies: List[float], ax: Axes) -> None:
        self.energies = energies
        self.ax = ax
        self.set_pretty()

    def update(self, image: int) -> None:
        self.vline.set_xdata([image, image])
        ymax = self.vline.get_ydata()[1]
        self.vline.set_ydata([self.energies[image], ymax])

    def plot(self) -> None:
        self.ax.plot(self.energies)

        ylims = self.ax.get_ylim()
        self.vline = self.ax.plot([0, 0], [self.energies[0], ylims[1]], "r--")[0]
        self.ax.set_ylim(ylims)

    def set_pretty(self) -> None:
        self.ax.set_xlabel("Image")
        self.ax.set_ylabel("Energy [eV]")
        self.ax.set_title("Energy plot")


@click.command(name="plot")
@click.argument("file", nargs=1, type=click.Path(exists=True))
@click.option("--output", "-o", type=click.Path(), help="Output file", default=None)
@click.option("--backend", "-b", type=str, help="Matplotlib backend.", default="TkAgg")
@click.option("--ase", "-a", is_flag=True, help="Use ASE to plot the structures.", default=False)
def cli_plot(file: str, output: Union[str, None], backend: str, ase: bool) -> None:
    """
    Plot structures from database or trajectory files.
    """
    from agox.analysis.search_analysis import SearchAnalysis
    from agox.databases.database_utilities import convert_database_to_traj

    matplotlib.use(backend)

    # Read the file:
    file = Path(file)
    if file.suffix == ".db":
        configurations = convert_database_to_traj(file)
    elif file.suffix == ".traj":
        configurations = read(file, ":")
    else:
        print("File type not")

    # # Create figure:
    if not ase:
        fig, axes = plt.subplots(1, 2, figsize=(10, 5))
        search_analysis = SearchAnalysis()
        search_analysis.plot_configuration(axes[0], candidate=configurations[0])

        energies = [config.get_potential_energy() for config in configurations]
        energy_plot = EnergyPlot(energies, axes[1])
        energy_plot.plot()

        # Create event handler:
        event_handler = EventHandler(search_analysis, {"configuration": axes[0]}, fig, configurations, energy_plot)
        fig.canvas.mpl_connect("key_press_event", event_handler.on_press)
        fig.canvas.mpl_connect("key_release_event", event_handler.on_release)

        if output is not None:
            plt.savefig(output)

        plt.show()
    else:
        view(configurations)



import logging
from pathlib import Path

import numpy as np
import pytorch_lightning as pl
import schnetpack as spk
import schnetpack.transform as trn
import torch
import torchmetrics
from ase import Atoms
from ase.calculators.calculator import Calculator, all_changes
from ase.calculators.singlepoint import SinglePointCalculator as SPC
from pytorch_lightning import seed_everything
from schnetpack.data import ASEAtomsData, AtomsDataModule
from schnetpack.interfaces import AtomsConverter
from schnetpack.representation import PaiNN

from agox.models.ABC_model import ModelBaseClass
from agox.observer import Observer

logging.getLogger("pytorch_lightning").setLevel(logging.ERROR)


class SchNetPackModel(ModelBaseClass):
    name = "SchNetPack-model"

    implemented_properties = ["energy", "forces"]

    dynamic_attributes = ["nnpot"]

    """ SchNetPack PaiNN model

    Parameters
    ----------
        
    """

    def __init__(
        self,
        cutoff=6.0,
        dataset_settings={},
        representation_settings={},
        loss_settings={},
        learning_settings={},
        trainer_callbacks=[],
        max_steps_per_iteration=100,
        max_epochs_per_iteration=10,
        representation_cls=PaiNN,
        base_path="",
        training_device=None,
        prediction_device=None,
        db_name="dataset.db",
        transfer_data=None,
        tensorboard=True,
        seed=None,
        **kwargs,
    ):
        super().__init__(**kwargs)

        self.tensorboard = tensorboard
        if seed is not None:
            seed_everything(seed, workers=True)

        if training_device is None:
            self.training_device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        else:
            self.training_device = torch.device(training_device)

        if prediction_device is None:
            self.prediction_device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        else:
            self.prediction_device = torch.device(prediction_device)

        self.base_path = Path(base_path)

        VERSION_NUMBER = 0
        while True:
            self.version = Path(f"version_{VERSION_NUMBER}_seed_{seed}")
            self.train_path = self.base_path / self.version
            if not self.train_path.is_dir():
                self.train_path.mkdir(parents=True, exist_ok=True)
                break
            else:
                VERSION_NUMBER += 1

        self.db_name = Path(db_name)
        self.data_path = self.train_path / self.db_name

        self.trainer_iteration = 1

        self.cutoff = cutoff
        self.dataset_settings = {
            **dataset_settings,
            **{
                "batch_size": 16,
                "num_train": 0.8,
                "num_val": 0.2,
                "transforms": [trn.ASENeighborList(cutoff=cutoff), trn.CastTo32()],
                "num_workers": 8,
                "pin_memory": True,
                "split_file": None,  # str(self.train_path / Path('split.npz')),
            },
        }
        self.representation_settings = {
            **representation_settings,
            **{
                "n_atom_basis": 96,
                "n_interactions": 5,
                "radial_basis": spk.nn.GaussianRBF(n_rbf=20, cutoff=cutoff),
                "cutoff_fn": spk.nn.CosineCutoff(cutoff),
            },
        }
        self.loss_settings = {**loss_settings, **{"energy_weight": 0.01, "forces_weight": 0.99}}
        self.learning_settings = {
            **learning_settings,
            **{
                "optimizer_cls": torch.optim.AdamW,
                "optimizer_args": {"lr": 1e-4},
                "scheduler_cls": spk.train.ReduceLROnPlateau,
                "scheduler_args": {"factor": 0.5, "patience": 1000, "verbose": True},
                "scheduler_monitor": "val_loss",
            },
        }

        self.ckpt_path = self.train_path / "last.ckpt"
        self.callbacks = trainer_callbacks + [
            pl.callbacks.ModelCheckpoint(dirpath=str(self.train_path), filename="last", save_last=True)
        ]
        # self.callbacks = [
        #     spk.train.ModelCheckpoint(
        #         model_path=str(self.train_path / Path('best_inference_model')),
        #         monitor='val_loss',
        #         save_top_k=-1,
        #         save_last=True,
        #         every_n_epochs=1,
        #     ),
        #     pl.callbacks.LearningRateMonitor(logging_interval='epoch')
        # ]

        self.max_steps_per_iteration = max_steps_per_iteration
        self.max_epochs_per_iteration = max_epochs_per_iteration
        self.representation_cls = representation_cls

        # Training DB
        if self.data_path.is_file():
            self.writer("ASE Database already exist. \n Connecting to existing database.")
            self.spk_database = ASEAtomsData(datapath=str(self.data_path))
        else:
            self.spk_database = ASEAtomsData.create(
                str(self.data_path), distance_unit="Ang", property_unit_dict={"energy": "eV", "forces": "eV/Ang"}
            )

        # Transfer data
        self.transfer_data = transfer_data
        if self.transfer_data is not None:
            self.add_data(self.transfer_data)

        # Model
        representation = self.representation_cls(**self.representation_settings)
        pred_energy = spk.atomistic.Atomwise(n_in=self.representation_settings.get("n_atom_basis"), output_key="energy")
        pred_forces = spk.atomistic.Forces(energy_key="energy", force_key="forces")

        pairwise_distance = spk.atomistic.PairwiseDistances()

        self.nnpot = spk.model.NeuralNetworkPotential(
            representation=representation,
            input_modules=[pairwise_distance],
            output_modules=[pred_energy, pred_forces],
            postprocessors=[
                trn.CastTo64(),
            ],
        )

        # Output
        output_energy = spk.task.ModelOutput(
            name="energy",
            loss_fn=torch.nn.MSELoss(),
            loss_weight=self.loss_settings.get("energy_weight"),
            metrics={"MAE": torchmetrics.MeanAbsoluteError()},
        )

        output_forces = spk.task.ModelOutput(
            name="forces",
            loss_fn=torch.nn.MSELoss(),
            loss_weight=self.loss_settings.get("forces_weight"),
            metrics={"MAE": torchmetrics.MeanAbsoluteError()},
        )

        self.task = spk.task.AtomisticTask(
            model=self.nnpot,
            outputs=[output_energy, output_forces],
            **self.learning_settings,
        )

        # Logging
        if self.tensorboard:
            self.logger = pl.loggers.TensorBoardLogger(
                save_dir=str(self.base_path), name=None, version=str(self.version)
            )
        else:
            self.logger = False

        ########## FOR PREDICTION ##########
        self.converter = AtomsConverter(
            neighbor_list=trn.ASENeighborList(cutoff=self.cutoff),
            device=self.prediction_device.type,
            dtype=torch.float32,
        )

        self.energy_key = "energy"
        self.forces_key = "forces"

    @property
    def transfer_data(self):
        return self._transfer_data

    @transfer_data.setter
    def transfer_data(self, l):
        if isinstance(l, list):
            self._transfer_data = l
            self._transfer_weights = np.ones(len(l))
        elif isinstance(l, dict):
            self._transfer_data = []
            self._transfer_weights = np.array([])
            for key, val in l.items():
                self._transfer_data += val
                self._transfer_weights = np.hstack((self._transfer_weights, float(key) * np.ones(len(val))))
        else:
            self._transfer_data = []
            self._trasfer_weights = np.array([])

    @property
    def transfer_weights(self):
        return self._transfer_weights

    def set_verbosity(self, verbose):
        self.verbose = verbose

    def calculate(self, atoms=None, properties=["energy", "forces"], system_changes=all_changes):
        Calculator.calculate(self, atoms, properties, system_changes)

        self.nnpot.eval()
        self.nnpot.to(device=self.prediction_device.type)  # , dtype=torch.float64)

        # Something inside schnetpack is checking for atoms objects so converting here.
        atoms = Atoms(numbers=atoms.get_atomic_numbers(), positions=atoms.positions, cell=atoms.cell, pbc=atoms.pbc)

        model_inputs = self.converter(atoms)
        model_results = self.nnpot(model_inputs)
        if "energy" in properties:
            e = model_results["energy"].cpu().data.numpy()[0].astype(np.float64)
            # print('Energy pred:', type(e), e, flush=True)
            self.results["energy"] = e

        if "forces" in properties:
            f = model_results["forces"].cpu().data.numpy().astype(np.float64)
            # print('Forces pred:', type(f), f.shape, flush=True)
            self.results["forces"] = f

    ####################################################################################################################
    # Prediction
    ####################################################################################################################

    def predict_energy(self, atoms=None, X=None, return_uncertainty=False):
        self.nnpot.eval()
        self.nnpot.to(device=self.prediction_device.type)

        model_inputs = self.converter(atoms)
        model_results = self.nnpot(model_inputs)

        return model_results["energy"].cpu().data.numpy()[0].astype(np.float64)

    def predict_energies(self, atoms_list):
        self.nnpot.eval()
        self.nnpot.to(device=self.prediction_device.type)

        model_inputs = self.converter(atoms_list)
        model_results = self.nnpot(model_inputs)

        return model_results["energy"].cpu().data.numpy().astype(np.float64)

    def predict_uncertainty(self, atoms=None, X=None, k=None):
        self.writer("Uncertainty not implemented.")
        return 0.0

    def predict_local_energy(self, atoms=None, X=None):
        self.writer("Local energy not implemented.")
        return np.zeros((len(atoms),))

    def predict_forces(self, atoms, return_uncertainty=False, **kwargs):
        self.nnpot.eval()
        self.nnpot.to(device=self.prediction_device.type)

        model_inputs = self.converter(atoms)
        model_results = self.nnpot(model_inputs)

        return model_results["forces"].cpu().data.numpy()

    def train_model(self, training_data, **kwargs):
        self.writer("Training PaiNN model")
        self.nnpot.train()
        self.nnpot.to(device=self.training_device.type)

        self.ready_state = True
        self.atoms = None

        self.add_data(training_data)

        # Dataloader
        dataset = AtomsDataModule(self.data_path, **self.dataset_settings)
        dataset.prepare_data()
        dataset.setup()

        trainer = pl.Trainer(
            accelerator=self.training_device.type,
            # devices=1,
            callbacks=self.callbacks,
            logger=self.logger,
            default_root_dir=str(self.train_path),
            max_epochs=self.trainer_iteration * self.max_epochs_per_iteration,
            max_steps=self.trainer_iteration * self.max_steps_per_iteration,
            enable_progress_bar=False,
            enable_model_summary=False,
        )
        ckpt_path = str(self.ckpt_path) if self.trainer_iteration > 1 else None
        trainer.fit(self.task, datamodule=dataset, ckpt_path=ckpt_path)

        self.trainer_iteration += 1

    def update_model(self, new_data, all_data):
        self.add_data(new_data)
        self.train_model([])

    ####################################################################################################################
    # Assignments:
    ####################################################################################################################

    @Observer.observer_method
    def training_observer(self, database, state):
        iteration = state.get_iteration_counter()

        if iteration < self.iteration_start_training:
            return
        if (iteration % self.update_period != 0) * (iteration != self.iteration_start_training):
            return

        all_data = database.get_all_candidates()
        self.writer(f"length all data: {len(all_data)}")

        if self.ready_state:
            full_update = False
            data_amount_before = len(self.spk_database) - len(self.transfer_data)
            data_for_training = all_data
            data_amount_new = len(data_for_training) - data_amount_before
            new_data = data_for_training[-data_amount_new:]
        else:
            full_update = True
            data_for_training = all_data

        if full_update:
            self.train_model(data_for_training)
        else:
            self.update_model(new_data, data_for_training)

    def load(self, path):
        state_dict = torch.load(path, map_location=self.prediction_device.type).state_dict()
        self.nnpot.load_state_dict(state_dict)

    def add_data(self, data_list):
        if len(data_list) == 0:
            return

        property_list = []
        for a in data_list:
            e = a.get_potential_energy(apply_constraint=False)
            f = a.get_forces(apply_constraint=False).reshape(-1, 3)
            c = SPC(a, energy=e, forces=f)
            a.set_calculator(c)
            properties = {"energy": np.array([e]), "forces": f}
            property_list.append(properties)
        self.spk_database.add_systems(property_list, data_list)



from typing import List, Union

import numpy as np
from matplotlib.axes import Axes

from agox.analysis.property import Property
from agox.analysis.search_data import SearchCollection, SearchData


class PropertyPlotter:
    def __init__(self, searches: Union[SearchCollection, List[SearchData]], search_property: Property, window: int = 5) -> None:
        """
        Plot an analysis property versus time for a list of searches. The property
        should be a scalar property, e.g. one number pr. iteration/time-step of a search.
        """
        self.searches = searches
        self.search_property = search_property
        self.window = window

    def plot(self, ax: Axes) -> None:
        for search in self.searches:
            self.plot_search(ax, search)

        ax.set_xlim(0, None)
        ax.legend()

    def plot_search(self, ax: Axes, search: SearchData) -> None:
        # Gather stuff
        prop = self.search_property(search)
        values = prop.data
        time = prop.axis[1]
        min_energies = np.minimum.accumulate(values, axis=1)

        time_mean = np.nanmean(time, axis=0)
        prop_mean = np.nanmean(values, axis=0)
        prop_min = np.nanmin(min_energies, axis=0)

        prop_min_avg = np.nanmean(min_energies, axis=0)

        roll_kernel = np.ones(self.window) / self.window
        roll_avg = np.convolve(prop_mean, roll_kernel, mode="valid")

        values_max = np.nanmax(values, axis=0)
        values_min = np.nanmin(values, axis=0)
        values_min_roll = np.convolve(values_min, roll_kernel, mode="valid")
        values_max_roll = np.convolve(values_max, roll_kernel, mode="valid")

        # Plot - mean
        (l1,) = ax.plot(time_mean, prop_min, zorder=1) # Minimum energy among all restarts
        ax.plot(time_mean, prop_min_avg, '--', color=l1.get_color(), zorder=1) # Average minimum energy among all restarts
        l2, = ax.plot(time_mean[:len(roll_avg)], roll_avg, label=f"{search.get_label()}", color=l1.get_color(), zorder=0, alpha=0.75) # Rolling average
        ax.fill_between(time_mean[:len(values_max_roll)], values_max_roll, values_min_roll, alpha=0.2, color=l1.get_color(), zorder=0.5) # Rolling average spread

        ax.set_xlabel(f"{prop.shape[1]}")
        ax.set_ylabel(f"{prop.name}")



# Configuration file for the Sphinx documentation builder.
import importlib
import inspect
import os


# -- Project information -----------------------------------------------------

project = "AGOX"
copyright = "2022-2025, AGOX Contributors"
author = "Mads-Peter V. C"

# from agox.__version__ import __version__ as agox_version
# The full version, including alpha/beta/rc tags
# Version Number:
import re
from agox import __version__

release = __version__

# -- Prolog ------------------------------------------------------------------

rst_prolog = """
.. |CHGNet| replace:: Can easily run with CHGNet using the calculator provided by the chgnet package (``pip install chgnet``). See https://chgnet.lbl.gov/ for more info.
.. |EMT| replace:: EMT is a quick potential that is bundled with ASE, not accurate enough for material science, but useful testing search algorithms.
.. |GPAW| replace:: Tested with GPAW 23.9.1 and ASE 3.25.0. See https://wiki.fysik.dtu.dk/gpaw/ for more info.
.. |ORCA| replace:: Tested with ORCA 5.0.4 and ASE 3.23.0, see https://wiki.fysik.dtu.dk/ase/ase/calculators/orca.html and https://sites.google.com/site/orcainputlibrary for more info.
.. |VASP| replace:: Tested with VASP 5.4.4 and ASE 3.23.0, see https://wiki.fysik.dtu.dk/ase/ase/calculators/vasp.html for more information. As of Feb 2025 VASP calculations done with ASE do not raise an exception when not the SCF does not converge, to avoid adding unphysical structures to the database the callback function described here :ref:`EVAL_CALLBACK` is used. 
"""

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.duration",
    "sphinx_tabs.tabs",
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "autoapi.extension",
    "sphinx.ext.linkcode",
    # 'sphinx.ext.autosectionlabel',
    # 'sphinx.ext.graphviz',
    # 'sphinx.ext.inheritance_diagram',
    "sphinx.ext.doctest",
    "sphinx_copybutton",
    "sphinx_design",
]

# Document Python Code
autoapi_type = "python"
autoapi_dirs = ["../../agox/"]
autoapi_ignore = ["*/*test*"]
autodoc_typehints = "description"
autoapi_keep_files = False
autoapi_options = [
    "members",
    "undoc-members",
    "private-members",
    "show-inheritance",
    "show-module-summary",
    "special-members",
    "imported-members",
]
autoapi_add_toctree_entry = False

autoapi_template_dir = "../autoapi_templates"

# Limk code:
code_url = "https://gitlab.com/agox/agox/-/blob/dev/{}?ref_type=heads#L{}-L{}"


def linkcode_resolve(domain, info):
    # Non-linkable objects from the starter kit in the tutorial.
    if domain == "js" or info["module"] == "connect4":
        return

    assert domain == "py", "expected only Python objects"

    try:
        mod = importlib.import_module(info["module"])
        if "." in info["fullname"]:
            objname, attrname = info["fullname"].split(".")
            obj = getattr(mod, objname)
            try:
                # object is a method of a class
                obj = getattr(obj, attrname)
            except AttributeError:
                # object is an attribute of a class
                return None
        else:
            obj = getattr(mod, info["fullname"])

        try:
            file = inspect.getsourcefile(obj)
            lines = inspect.getsourcelines(obj)
        except TypeError:
            # e.g. object is a typing.Union
            return None
        except OSError:  # Cytohn
            return None
        except:  # other issues
            return None

        file = os.path.relpath(file, os.path.abspath(".."))

        start, end = lines[1], lines[1] + len(lines[0]) - 1

        return code_url.format(file, start, end)
    except:  # The mother of catching all exceptions
        return None


# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "furo"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

html_favicon = "A.ico"

sd_custom_directives = {
    'script-tab-set': {
        'inherit': 'tab-set',
        'options':
            {
                "class": "sd-pl-1 sd-ml-3 sd-rounded-2"
            }
    }
}



import rich_click as click


@click.command("notebook")
@click.argument("directories", nargs=-1, type=str)
@click.option("--auto-open", is_flag=True, help="Open the notebook in VS-code after creation.")
@click.option("--name", type=str, help="Name of the notebook file.")
def cli_notebook(directories, auto_open, name) -> None:
    """
    Create a jupyter notebook for analysis.
    """
    import pathlib
    import subprocess

    import nbformat as nbf

    # Need to massage the input of directories a little bit:
    for i in range(len(directories)):
        directories[i] = "'" + directories[i] + "'"

    # Convenience settings:
    figsize = "figsize = (7, 7)"

    # Create notebook object
    nb = nbf.v4.new_notebook()

    # Add functions:
    markdown = nbf.v4.new_markdown_cell
    code = nbf.v4.new_code_cell

    # List to hold cell contents
    cells = nb["cells"] = []

    def add_cell(content_string, add_function):
        cells.append(add_function(content_string))

    # Create initial text
    this_file = pathlib.Path(__file__).absolute()
    headline = """
# Automagic AGOX Analysis
This is an automatically generated AGOX analysis notebook.\n 
Notebook generation is defined in this file: {}
    """.format(this_file)
    # cells.append(markdown(headline))
    add_cell(headline, markdown)

    # Import statements % basic settings
    import_block = """import numpy as np
import ipywidgets as widgets
import matplotlib.pyplot as plt
from agox.utils.batch_analysis import Analysis
from agox.utils.jupyter_interactive import InteractiveStructureHistogram, InteractiveSuccessStats, InteractiveEnergy, sorted_species"""
    # cells.append(code(import_block))
    add_cell(import_block, code)

    # Load block:
    load_headline = """## Scan and read db-files in the directories and calculate CDF.
This is the block that takes the most times, avoid rerunning unless changes have been made that require re-reading the 
directories, such as adding additional ones. """
    add_cell(load_headline, markdown)

    load_block = """
analysis = Analysis()
force_reload = False
"""
    for directory in directories:
        load_block += "analysis.add_directory({}, force_reload=force_reload)\n".format(directory)
    load_block += "analysis.compile_information()\n"
    load_block += "analysis.calculate_CDF()"

    # cells.append(code(load_block))
    add_cell(load_block, code)

    # Success statistics block:
    succes_headline = """
## Analysis
    """
    add_cell(succes_headline, markdown)

    success_block = """ISS = InteractiveSuccessStats(analysis)
out = widgets.interactive_output(ISS.update_plot, ISS.widget_dict)
widgets.HBox([widgets.VBox(ISS.widget_list), out])""".format()
    add_cell(success_block, code)

    energy_int_block = """IE = InteractiveEnergy(analysis)
out = widgets.interactive_output(IE.update_plot, IE.widget_dict)
widgets.HBox([widgets.VBox(IE.widget_list), out])"""
    add_cell(energy_int_block, code)

    interactive_struct_hist = """ISH = InteractiveStructureHistogram(analysis)
num_structures = np.sum(analysis.restarts)
index = widgets.IntSlider(min=0, max=num_structures-1, value=0, description='Index')
widgets.interactive(ISH.update_plot, index=index)"""
    add_cell(interactive_struct_hist, code)

    structure_viewer = """from ase.visualize import view
structures, energies = analysis.get_best_structures()
structures = [sorted_species(atoms) for atoms in structures]
view(structures, viewer='ngl')"""
    add_cell(structure_viewer, code)

    # Write notebook:
    notebook_path = name + ".ipynb"
    nbf.write(nb, notebook_path)

    # This will only work with appropriately configured VS-code installation.
    if auto_open:
        subprocess.run("code {}".format(notebook_path), shell=True)



import matplotlib

matplotlib.use("Agg")

import numpy as np
from agox import AGOX
from agox.databases import Database
from agox.environments import Environment
from agox.evaluators import LocalOptimizationEvaluator
from agox.generators import RandomGenerator
from ase import Atoms

##############################################################################
# Calculator
##############################################################################

from ase.calculators.emt import EMT

calc = EMT()

##############################################################################
# General settings:
##############################################################################

template = Atoms("", cell=np.eye(3) * 12)
confinement_cell = np.eye(3) * 6
confinement_corner = np.array([3, 3, 3])
environment = Environment(
    template=template, symbols="Au5Ni", confinement_cell=confinement_cell, confinement_corner=confinement_corner
)

# Database
db_path = "db1.db"
database = Database(filename=db_path, order=3)

##############################################################################
# Search Settings:
##############################################################################

random_generator = RandomGenerator(**environment.get_confinement(), environment=environment, order=1)

evaluator = LocalOptimizationEvaluator(
    calc,
    gets={"get_key": "candidates"},
    optimizer_kwargs={"logfile": None},
    store_trajectory=False,
    optimizer_run_kwargs={"fmax": 0.05, "steps": 400},
    order=2,
)

##############################################################################
# Let get the show running!
##############################################################################

from agox.observer import Observer


class TwoMethodObserver(Observer):
    name = "TwoMethodObserver"

    def __init__(self, order=[2, 4], gets=[{"get_generated": "candidates"}, {"get_evaluated": "evaluated_candidates"}]):
        Observer.__init__(self, gets=gets, order=order, sets={})
        self.add_observer_method(self.observer_method_1, order=self.order[0], sets=self.sets[0], gets=self.gets[0])
        self.add_observer_method(self.observer_method_2, order=self.order[1], sets=self.sets[0], gets=self.gets[1])

        self.mean = 0.5

    @Observer.observer_method
    def observer_method_1(self, state):
        self.writer(f"Iteration: {self.get_iteration_counter()}: I AM A MODULE.")

        candidates = state.get_from_cache(self, self.get_generated)

        # Very advanced prediction algorithm:
        # 1. First look at the candidates.
        number = 0
        for candidate in candidates:
            number += np.sum(candidate.positions)

        # 2. Then choose a random number regardless of that.
        chance = np.random.normal(loc=self.mean, scale=1, size=1)

        if chance > 0.5:
            self.prediction = "Good"
        else:
            self.prediction = "Bad"

        self.writer(f"I predict the candidate will be: {self.prediction}")

    @Observer.observer_method
    def observer_method_2(self, state):
        candidates = state.get_from_cache(self, self.get_evaluated)

        for candidate in candidates:
            E = candidate.get_potential_energy()

            self.writer(f"I predicted candidate would be {self.prediction} and it has energy E = {E}.")


basic_observer = TwoMethodObserver()

agox = AGOX(random_generator, database, basic_observer, evaluator)

agox.run(N_iterations=1)



import json
from pathlib import Path
from typing import Dict, Optional, Union

import numpy as np
from ase.data import atomic_numbers

symbols = {n: s for s, n in atomic_numbers.items()}


class ThermodynamicsData:
    def __init__(
        self,
        references: Dict[str, float],
        chemical_potentials: Dict[str, float],
        template_energy: Optional[float] = 0.0,
    ) -> None:
        """
        Container for thermodynamic data.

        Parameters
        ----------
        references :
            Reference energies for each element. Keys are element symbols.
        chemical_potentials :
            Chemical potentials for each element. Keys are element symbols.
        template_energy :
            Energy of the template structure.
        """
        self.references = references
        self.chemical_potentials = chemical_potentials
        self.template_energy = template_energy

    def set_chemical_potential(self, element: Union[str, int], value: float) -> None:
        """
        Set the chemical potential of an element.

        Parameters
        ----------
        element : str or int
            Element symbol or atomic number.
        value : float
            Chemical potential.
        """
        element = self.convert_number_to_element(element)
        self.chemical_potentials[element] = value

    def get_chemical_potential(self, element: Union[str, int]) -> float:
        """
        Get the chemical potential of an element.

        Parameters
        ----------
        element : str or int
            Element symbol or atomic number.
        """
        element = self.convert_number_to_element(element)
        return self.chemical_potentials[element]

    def set_reference(self, element: str, value: float) -> None:
        """
        Set the reference energy of an element.

        Parameters
        ----------
        element : str or int
            Element symbol or atomic number.
        value : float
            Reference energy.
        """
        element = self.convert_number_to_element(element)
        self.references[element] = value

    def get_reference(self, element: Union[str, int]) -> float:
        """
        Get the reference energy of an element.

        Parameters
        ----------
        element : str or int
            Element symbol or atomic number.
        """
        element = self.convert_number_to_element(element)
        return self.references[element]

    def convert_number_to_element(self, element: Union[int, str]) -> str:
        """
        Convert an atomic number to an element symbol.

        Parameters
        ----------
        element : int or str
            Atomic number or element symbol.

        Returns
        -------
        str
            Element symbol.
        """
        if isinstance(element, (int, np.int64)):
            return symbols[element]
        return element

    @classmethod
    def from_file(cls, path: Union[str, Path]):
        """
        Load thermodynamic data from a file.

        Parameters
        ----------
        path : str or Path
            Path to the file.

        Returns
        -------
        ThermodynamicsData
            Thermodynamic data.
        """
        if isinstance(path, str):
            path = Path(path)

        if not path.exists():
            raise FileNotFoundError(f"{path} does not exist")

        with open(path, "r") as f:
            data = json.load(f)

        return cls(
            references=data["references"],
            chemical_potentials=data["chemical_potentials"],
            template_energy=data.get("template_energy", 0.0),
        )
    
    @classmethod
    def load(cls, path: Path):
        return ThermodynamicsData.from_file(path)

    def save(self, path: Union[str, Path], overwrite: bool = False) -> None:
        """
        Save thermodynamic data to a file.

        Parameters
        ----------
        path : str or Path
            Path to the file.
        overwrite : bool
            Whether to overwrite the file if it already exists.
        """
        if isinstance(path, str):
            path = Path(path)

        data = {
            "references": self.references,
            "chemical_potentials": self.chemical_potentials,
            "template_energy": self.template_energy,
        }

        if not path.exists() or overwrite:
            with open(path, "w") as f:
                json.dump(data, f, indent=4)
        else:
            raise FileExistsError(f"{path} already exists")


from typing import Literal

import numpy as np

from agox.analysis.search_data import SearchData

from .property import ArrayPropertyData, Property


class EnergyProperty(Property):
                    
    def compute(self, search_data: SearchData) -> np.array:
        """
        Get the energy of the system as a np.array of shape [restarts, iterations].
        """

        axis_name, indices = self.get_time_axis(search_data)

        energy = ArrayPropertyData(
            data=search_data.get_all("energies", fill=np.inf),
            name="Energy",
            shape=("Restarts", axis_name),
            array_axis=(search_data.get_all_identifiers(), indices),
        )
        return energy



import numpy as np
from scipy.optimize import minimize

from agox.generators.ABC_generator import GeneratorBaseClass


class ComplementaryEnergyGenerator(GeneratorBaseClass):
    """
    Use complementary energy landscape to generate new candidates.

    Parameters
    ------------
    calc: CE calculator
        Used to evaluate the CE energy and forces
    descriptor: Local descriptor
        Used to calculate local descriptor for atoms in structures
    attractor_method: Method for picking attractors
        Used to construct attractors in CE expression
    move_all: Bool
        Whether to move all non-template atoms or not. If "mover_indices"
        is not None, the atoms with the corresponding indices will be free
        to move, thereby ignoring the "move_all" argument.
    mover_indices: List of integers
        Only atoms with the corresponding indices can move during
        candidate generation.
    """

    name = "ComplementaryEnergyGenerator"

    def __init__(
        self, calculator, descriptor, attractor_method, move_all=True, mover_indices=None, replace=True, **kwargs
    ):
        super().__init__(replace=replace, **kwargs)
        self.calc = calculator
        self.descriptor = descriptor
        self.attractor_method = attractor_method
        self.move_all = move_all
        self.mover_indices = mover_indices

    def _get_candidates(self, candidate, parents, environment):
        template = candidate.get_template()
        n_template = len(template)

        # Set attractors
        attractors = self.attractor_method.get_attractors(candidate)
        self.calc.attractors = attractors

        # Determine which atoms to move if no indices has been supplied
        # Picks those with highest local complementary energy
        if self.mover_indices is None:
            if self.move_all:
                mover_indices = range(n_template, len(candidate))
            else:
                n_movers = np.random.randint(1, len(candidate) - n_template)
                mover_indices = np.array(range(n_template, len(candidate)))
                self.calc.mover_indices = mover_indices
                local_ce = self.calc.get_local_ce_energy(candidate)
                largest_local_ce_indices = np.argsort(local_ce)[::-1]
                mover_indices = mover_indices[largest_local_ce_indices][:n_movers]

            self.mover_indices = mover_indices

        # Set mover indices and attractors for calculator
        self.calc.mover_indices = self.mover_indices

        # Set calculator
        candidate.calc = self.calc

        def objective_func(pos):
            candidate.positions[self.mover_indices] = np.reshape(pos, [len(self.mover_indices), 3])
            ce = candidate.get_potential_energy(apply_constraint=0)
            grad = -candidate.get_forces(apply_constraint=0)[self.mover_indices]
            return ce, grad.flatten()

        pos = candidate.positions[self.mover_indices].flatten()
        pos = minimize(objective_func, pos, method="BFGS", jac=True, options={"maxiter": 75, "disp": False}).x
        suggested_positions = np.reshape(pos, [len(self.mover_indices), 3])

        for index in range(len(self.mover_indices)):
            i = self.mover_indices[index]
            for _ in range(100):
                if _ == 0:
                    radius = 0
                else:
                    radius = 0.5 * np.random.rand() ** (1 / self.get_dimensionality())
                displacement = self.get_displacement_vector(radius)
                suggested_position = suggested_positions[index] + displacement

                # Check confinement limits:
                if not self.check_confinement(suggested_position):
                    continue

                # Check that suggested_position is not too close/far to/from other atoms
                # Skips the atom it self.
                if self.check_new_position(candidate, suggested_position, candidate[i].number, skipped_indices=[i]):
                    candidate[i].position = suggested_position
                    break

        return [candidate]

    @classmethod
    def default(
        cls,
        environment,
        database,
        move_all=True,
        mover_indices=None,
        attractors_from_template=False,
        predefined_attractors=None,
        **kwargs,
    ):
        from agox.generators.complementary_energy.attractor_methods.ce_attractors_current_structure import (
            AttractorCurrentStructure,
        )
        from agox.generators.complementary_energy.ce_calculators import ComplementaryEnergyDistanceCalculator
        from agox.models.descriptors.exponential_density import ExponentialDensity

        lambs = [0.5, 1, 1.5]
        rc = 10.0
        descriptor = ExponentialDensity(environment=environment, lambs=lambs, rc=rc)
        ce_calc = ComplementaryEnergyDistanceCalculator(descriptor=descriptor)
        ce_attractors = AttractorCurrentStructure(
            descriptor=descriptor,
            attractors_from_template=attractors_from_template,
            predefined_attractors=predefined_attractors,
        )
        ce_attractors.attach(database)

        return cls(
            calculator=ce_calc,
            descriptor=descriptor,
            attractor_method=ce_attractors,
            move_all=move_all,
            **environment.get_confinement(),
        )

    def get_number_of_parents(self, sampler):
        return 1



from typing import Optional

import numpy as np
from ase import Atoms

from agox.models.ABC_model import ModelBaseClass


class CompositionModel(ModelBaseClass):
    name = "CompositionModel"
    implemented_properties = ["energy", "forces"]

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self.atom_energies = np.zeros(100)
        self.add_dynamic_attribute("atom_energies")

    def predict_energy(self, atoms: Atoms, **kwargs) -> float:
        numbers = atoms.get_atomic_numbers()
        return np.sum(self.atom_energies[numbers])

    def predict_forces(self, atoms: Atoms, **kwargs) -> np.ndarray:
        return np.zeros((len(atoms), 3))
    
    def reset(self) -> None:
        self.atom_energies = np.zeros(100)

    def train(self, data: list[Atoms], energies: Optional[np.ndarray] = None) -> None:
        self.atom_energies = np.zeros(100)
        n = len(self.atom_energies)
        features = np.zeros((len(data), n))

        for i, atoms in enumerate(data):
            numbers = atoms.get_atomic_numbers()
            for number in np.unique(numbers):
                features[i, number] += np.count_nonzero(numbers == number)

        if energies is None:
            energies = np.array([atoms.get_potential_energy() for atoms in data])

        self.atom_energies = np.linalg.pinv(features.T @ features) @ features.T @ energies



from agox.generators.ABC_generator import GeneratorBaseClass
from agox.candidates.standard import StandardCandidate

from ase import Atoms, Atom
import numpy as np

from copy import deepcopy as dcpy


class AddRemoveGenerator(GeneratorBaseClass):
    name = "AdditionRemovalGenerator"

    def __init__(self, max_diff=0, **kwargs):
        super().__init__(**kwargs)
        """Maximum difference of additions and removals from the initial stoichiometry and the final one, 
        <1 means any.
        """
        self.max_diff = max_diff

    def get_composition_delta(self, a1: Atoms, a2: Atoms) -> np.array:
        u_elements = np.unique(np.append(a1.numbers, a2.numbers))
        differences = [
            int((a2.numbers == u).sum() - (a1.numbers == u).sum()) for u in u_elements
        ]

        add_rm = []
        for e, d in zip(u_elements, differences):
            add_rm.extend([e * np.sign(d)] * abs(d))
        return add_rm

    def _get_candidates(self, candidate, parents, environment):
        template = environment.get_template()

        numbers_list_array = environment.get_numbers_list()
        numbers_list = numbers_list_array[np.random.randint(0, len(numbers_list_array))]
            
        a1 = parents[0][len(template) :] # Get the non-template atoms.
        a2 = Atoms(numbers=numbers_list) # Get a random composition.

        # Get the difference in composition.
        delta = self.get_composition_delta(a1, a2)

        t = 0
        while t < 100:
            t + 1
            new_atoms = dcpy(template + a1)

            delta_rand = np.random.permutation(delta)


            success = True
            for d in delta_rand:
                if d > 0:
                    new_atoms, s = self.add_atom(new_atoms, d)
                    if not s:
                        success = False
                        break
                if d < 0:
                    new_atoms, s = self.remove_atom(template, new_atoms, -d)
                    if not s:
                        success = False
                        break
            if success:
                return [StandardCandidate.from_atoms(template, new_atoms)]

        self.writer("AddRemoveGenarator failing at producing valid structure")
        return []

    def add_atom(
        self,
        atoms: Atoms,
        element: int,
    ) -> tuple:  # [Atoms,bool]:
        if len(atoms) == 0:
            p = self.get_box_vector()

            return Atoms(numbers=[element], positions=[p], cell=atoms.cell), True
        t = 0
        new_atoms = dcpy(atoms)
        while t < 100:
            idx = np.random.randint(len(atoms))
            placed_atom = atoms[idx]
            suggested_position = placed_atom.position.copy()
            # Get a vector at an appropriate radius from the picked atom.
            vec = self.get_sphere_vector(element, placed_atom.number)
            suggested_position += vec

            if not self.check_confinement(suggested_position).all():
                continue
            if (
                self.check_new_position(atoms, suggested_position, element)
                or len(atoms) == 0
            ):
                new_atoms.extend(
                    Atoms(numbers=[element], positions=[suggested_position])
                )
                return new_atoms, True

        return new_atoms, False

    def remove_atom(
        self,
        template: Atoms,
        atoms: Atoms,
        element: int,
    ) -> tuple:  # [Atoms,bool]:


        remove_pos = np.flatnonzero(atoms[len(template) :].numbers == element) + len(
            template
        )

        if len(remove_pos) == 0:
            return atoms, False

        rm = np.random.permutation(remove_pos)
        del atoms[rm[0]]
        return atoms, True

    def get_number_of_parents(self, sampler):
        return 1



from typing import Any, Tuple

import numpy as np
from scipy.spatial.distance import cdist

from agox.analysis.criterion.base_criterion import BaseCriterion
from agox.analysis.property.property import ListPropertyData


class DistanceCriterion(BaseCriterion):
    def __init__(self, threshold: float, comparate: Any) -> None:
        """
        Find the time at which the property first crosses below a threshold.

        Parameters
        ----------
        threshold : float
            Distance threshold to compare against the property data
        comparate : Any
            Data to compare against the property data, of the same type as the objects in property_data.data
        """
        self.threshold = threshold
        self.comparate = comparate

    def compute(self, property_data: ListPropertyData) -> Tuple[np.array, np.array]:
        """
        Parameters
        ----------
        property_data : ListPropertyData
            Property data to be analyzed
        """
        time_indices = []
        events = []

        for data in property_data.data:
            property_arr = np.array(data).squeeze().reshape(-1, self.comparate.shape[1])
            distances = cdist(property_arr, self.comparate)
            below_threshold = distances <= self.threshold
            if below_threshold.any():
                indices = np.argwhere(below_threshold).flatten()
                time = indices[0]
                event = 1
            else:
                time = len(property_arr)-1
                event = 0

            time_indices.append(time)
            events.append(event)

        events = np.array(events)
        times = self._get_times(property_data, time_indices, events)

        # times = np.array([property_data.axis[1][i, time_index] for i, time_index in enumerate(time_indices)])
        # times[events == 0] = np.nanmax(property_data.axis[1])

        print(f"Distance criterion: {np.sum(events)} events found")
        print(events)

        return times, events
    
    def _get_times(self, property_data, time_indices, events):
        times = np.array([property_data.axis[1][i, time_index] for i, time_index in enumerate(time_indices)])
        times[events == 0] = np.nanmax(property_data.axis[1])
        return times



import numpy as np
import matplotlib.pyplot as plt

from agox.models import load
from agox.models.datasets import datasets

data = datasets["Ag5O3"]
test_data = data[80:]

model = load("my-model.pkl")

true_energies = np.array([d.get_potential_energy() for d in test_data])
pred_energies = model.predict_energies(test_data)

# Make parity plot
fig, ax = plt.subplots()
ax.scatter(true_energies, pred_energies)

min_energy, max_energy, q = (
    np.min(np.vstack((true_energies, pred_energies))),
    np.max(np.vstack((true_energies, pred_energies))),
    0.2,
)
ax.plot([min_energy - q, max_energy + q], [min_energy - q, max_energy + q], "-k")
ax.set_xlim(([min_energy - q, max_energy + q]))
ax.set_ylim(([min_energy - q, max_energy + q]))
ax.set_xlabel("True Energies")
ax.set_ylabel("Predicted Energies")
ax.set_title(f"Test MAE: {np.mean(np.abs(true_energies-pred_energies)/len(data[0]))*1000:.2f} meV/atom")
plt.savefig("parity.png")




import numpy as np

from agox.acquisitors import LowerConfidenceBoundAcquisitor
from agox.utils.thermodynamics import ThermodynamicsData, gibbs_free_energy


class GibbsFreeEnergyAquisitor(LowerConfidenceBoundAcquisitor):
    name = "GFEacquisitor"

    def __init__(
        self,
        thermo_data: ThermodynamicsData,
        model: object,
        kappa: float = 1,
        **kwargs,
    ):
        super().__init__(model, kappa, **kwargs)
        self.thermo_data = thermo_data

    def calculate_acquisition_function(
        self, candidates: list
    ) -> np.array:
        fitness = np.zeros(len(candidates))

        # Attach calculator and get model_energy
        for i, candidate in enumerate(candidates):
            E, sigma = self.model.predict_energy_and_uncertainty(candidate)

            fitness[i] = gibbs_free_energy(candidate=candidate, total_energy=self.acquisition_function(E, sigma), 
                                           thermo_data=self.thermo_data)

            # For printing:
            comp = candidate.get_search_formula()

            ##------------------------------
            candidate.add_meta_information("model_energy", E)
            candidate.add_meta_information("uncertainty", sigma)
            candidate.add_meta_information("model_gfe", fitness[i])
            candidate.add_meta_information("gfe_composition", comp)

        return fitness

    def print_information(self, candidates, acquisition_values):
        if self.model.ready_state:
            for i, candidate in enumerate(candidates):
                fitness = acquisition_values[i]

                GFEmodel = candidate.get_meta_information("model_gfe")
                composition = candidate.get_meta_information("gfe_composition")
                Emodel = candidate.get_meta_information("model_energy")
                sigma = candidate.get_meta_information("uncertainty")
                description = candidate.get_meta_information("description")
                self.writer(
                    "Candidate "
                    + composition
                    + ": GFE={:8.3f}, E={:8.3f}, s={:8.3f}, F={:8.3f}, {}".format(
                        GFEmodel, Emodel, sigma, fitness, description
                    )
                )



"""
Plotting tools for analysis of AGOX searches.
"""

from agox.analysis.plot.property_plot import PropertyPlotter
from agox.analysis.plot.success_plot import SuccessPlotter

__all__ = ["SuccessPlotter", "PropertyPlotter"]



"""
A candidate is the basic piece of data being moved around in the algorithm.
Generators create candidates, and evaluators evaluate the objective function
at the coordinates described by the candidate.
"""
# ruff: noqa: I001, E402
from typing import TypeAlias

from agox.candidates.ABC_candidate import CandidateBaseClass
Candidate: TypeAlias = CandidateBaseClass

from agox.candidates.standard import StandardCandidate

__all__ = ["CandidateBaseClass", 
           "Candidate",
           "StandardCandidate"]



import os
from argparse import ArgumentParser

import numpy as np
from ase.io import write

from agox.databases.database import Database


def convert_database_to_traj(database_path):
    """
    Convert database to list of ASE Atoms objects.

    Currently assumes that the database can be read using the standard database
    class (agox/modules/databases/database.py).

    Parameters
    -----------
    database_path: str
        Path to database on disk.

    Returns
    --------
    list
        List of ASE Atoms objects.
    """

    if os.path.exists(database_path):
        database = Database(database_path, initialize=False)
    else:
        print(f"Trajectory not found: {database_path}")
        return []

    # trajectory = database.restore_to_trajectory()
    database.restore_to_memory()
    trajectory = database.get_all_candidates()

    for candidate in trajectory:
        candidate.info.update(candidate.meta_information)

    return trajectory


def find_file_types(paths):
    """
    Find file extension of given paths.

    Parameters
    -----------
    paths: list of str
        Paths to check

    Returns
    --------
    list:
        List of strs indicating of the file types (extensions) of the given paths.
    """
    file_types = []
    for path in paths:
        file_types.append(path.split(".")[-1])
    return file_types


def convert():
    """
    Command line function to convert database files to trajectories.

    Does not take any Python inputs but has the following comamnd line arguments:

    files: (positional): Paths of database files to convert e.g. given as db1.db db2.db
    name: (-n, --name): Name of the trajectory file to save including .traj extension.
    """
    parser = ArgumentParser()
    parser.add_argument("files", type=str, nargs="+")
    parser.add_argument("-n", "--name", type=str, default=None)
    args = parser.parse_args()

    # Determine the file extension & ensure all paths have the same extension.
    file_types = find_file_types(args.files)
    assert np.array([file_type == file_types[0] for file_type in file_types]).all()
    file_type = file_types[0]

    # If database files - convert to trajectory.
    trajectory = []
    if file_type == "db":
        for path in args.files:
            print(f"Converting: {path}")
            trajectory += convert_database_to_traj(path)

        # Naming:
        name = args.name
        if name is None:
            name = "converted_db.traj"

        # Save:
        print(f"Saving trajectory file {name}")
        write(name, trajectory)


if __name__ == "__main__":
    convert()



from collections import OrderedDict
from typing import Callable, Set, Tuple

import numpy as np

from agox.observer import ObserverMethod
from agox.writer import Writer


class ObserverHandler:
    """
    Base-class for classes that can have attached observers.
    """

    def __init__(self, handler_identifier: str, dispatch_method: Callable) -> None:
        """
        Parameters
        -----------
        handler_identifier: str
            Identifier for the handler.
        dispatch_method: Callable
            Method to dispatch to observers.
        """
        self.writer = Writer()
        self.observers = {}
        self.execution_sort_idx = []
        self.handler_identifier = handler_identifier
        self.set_dispatch_method(dispatch_method)

    ####################################################################################################################
    # General handling methods.
    ####################################################################################################################

    def attach_observer(self, observer_method: ObserverMethod) -> None:
        """
        Attaches an observer, saving it to the internal dict and re-evaluating
        the execution order.

        Parameters
        -----------
        observer_method: object
            An instance of the ObserverMethod class that will be attached.
        """
        self.observers[observer_method.key] = observer_method
        self.evaluate_execution_order()

    def delete_observer(self, observer_method: ObserverMethod) -> None:
        """
        Removes an observer, deleting it from the internal dict and re-evaluating
        the execution order.

        Parameters
        -----------
        observer_method: object
            An instance of the ObserverMethod class that will be attached.
        """
        del self.observers[observer_method.key]
        self.evaluate_execution_order()

    def evaluate_execution_order(self) -> None:
        """
        Evaluates the execution order by sorting the orders of stored
        observers.
        """
        keys = self.observers.keys()
        orders = [self.observers[key]["order"] for key in keys]
        self.execution_sort_idx = np.argsort(orders)

    def get_ordered_observers(self) -> OrderedDict:
        keys = [key for key in self.observers.keys()]
        observers = [self.observers[key] for key in keys]
        ordered_dict = OrderedDict()
        sort_idx = self.execution_sort_idx
        for index in sort_idx:
            ordered_dict[keys[index]] = observers[index]
        return ordered_dict

    def dispatch_to_observers(self, *args, **kwargs) -> None:
        """
        Dispatch to observers.

        Only rely on the order of execution if you have specified the 'order' argument for each observer.
        """
        observers = self.get_ordered_observers()
        for observer_method in observers.values():
            observer_method(*args, **kwargs)

    def set_dispatch_method(self, method: Callable) -> None:
        self.dispatch_method = self.name + "." + method.__name__

    ####################################################################################################################
    # Printing / Reporting
    ####################################################################################################################

    def print_observers(self, include_observer: bool = False, verbose: int = 0, hide_log: bool = True) -> None:
        """
        Parameters
        -----------
        include_observer: bool (default: False)
            Turns printing of observer name on. Might be useful to debug if
            something is not working as expected.
        verbose: int (default: 0)
            If verbose > 1 then more information is printed.
        hide_log: bool (default: True)
            If True then the LogEntry observers are not printed. Keeps the
            report more clean.
        """

        order_indexs = self.execution_sort_idx
        keys = [key for key in self.observers.keys()]
        names = [obs["name"] for obs in self.observers.values()]
        methods = [obs["method"] for obs in self.observers.values()]
        orders = [obs["order"] for obs in self.observers.values()]

        base_string = "{}: order = {} - name = {} - method - {}"
        if include_observer:
            base_string += " - method: {}"

        self.writer.write_header("Observers")
        for idx in order_indexs:
            self.writer("  Order {} - Name: {}".format(orders[idx], names[idx]))
            if verbose > 1:
                self.writer("  Key: {}".format(keys[idx]))
                self.writer("  Method: {}".format(methods[idx]))
                self.writer("_" * 50)

    def observer_reports(self, report_key: bool = False, hide_log: bool = True) -> None:
        """
        Generate observer report, which checks if the data flow is valid.

        Parameters
        -----------
        report_key: bool
            Whether or not to print the keys used to get or set from the cache.
        hide_log: bool
            Whether or not to print the LogEntry observers.
        """
        dicts_out_of_order = [value for value in self.observers.values()]

        self.writer.write_header("Observers set/get reports")

        base_offset = "  "
        extra_offset = base_offset + "    "
        for i in self.execution_sort_idx:
            observer_method = dicts_out_of_order[i]

            self.writer(base_offset + observer_method.name)
            report = observer_method.report(
                offset=extra_offset, report_key=report_key, print_report=False, return_report=True
            )
            for string in report:
                self.writer(string)

        get_set, set_set = self.get_set_match()
        self.writer(base_offset)
        self.writer(base_offset + "Overall:")
        self.writer(base_offset + f"Get keys: {get_set}")
        self.writer(base_offset + f"Set keys: {set_set}")
        self.writer(base_offset + f"Key match: {get_set==set_set}")
        if not get_set == set_set:
            self.writer(base_offset + "Sets do not match, this can be problematic!")
            if len(get_set) > len(set_set):
                self.writer(base_offset + "Automatic check shows observers will attempt to get un-set item!")
                self.writer(base_offset + "Program likely to crash!")
            if len(set_set) > len(get_set):
                self.writer(base_offset + "Automatic check shows observers set value that is unused!")
                self.writer(base_offset + "May cause unintended behaviour!")

            unmatched_keys = list(get_set.difference(set_set)) + list(set_set.difference(get_set))
            self.writer(base_offset + f"Umatched keys {unmatched_keys}")

    def get_set_match(self) -> Tuple[Set, Set]:
        """
        Check if gets and sets match.
        """
        dicts_out_of_order = [value for value in self.observers.values()]
        all_sets = []
        all_gets = []

        for observer_method in dicts_out_of_order:
            all_sets += observer_method.sets.values()
            all_gets += observer_method.gets.values()

        all_sets = set(all_sets)
        all_gets = set(all_gets)

        return all_gets, all_sets



import pickle
from copy import deepcopy
from dataclasses import dataclass
from pathlib import Path
from typing import List, Optional, Union
from uuid import uuid4

import numpy as np
from ase import Atoms
from rich.progress import track

from agox.candidates import StandardCandidate
from agox.databases import Database


def read_database(path: Path, database_class: Database) -> List[StandardCandidate]:
    try:
        db = database_class(path)
        db.restore_to_memory()
        candidates = db.get_all_candidates()
    except: # noqa: E722
        return False, []

    if len(candidates) == 0:
        return False, []

    return True, candidates


@dataclass
class RestartData:
    path: Path
    candidates: List[Atoms]
    energies: np.array = None
    indices: np.array = None
    iterations: np.array = None
    times: np.array = None
    identifier: str = None

    def __post_init__(self) -> None:
        self.n_candidates = len(self.candidates)
        self.indices = np.arange(self.n_candidates)
        self.iterations = np.array([candidate.get_meta_information("iteration") for candidate in self.candidates])
        self.times = np.array([candidate.get_meta_information("time_stamp") for candidate in self.candidates])
        if self.energies is None:
            self.energies = np.array([candidate.get_potential_energy() for candidate in self.candidates])

        if self.identifier is None:
            self.identifier = self.path.stem

    def __len__(self) -> int:
        return self.n_candidates


class SearchData:
    """
    A set of restarts with the same settings.

    Parameters
    ----------
    directory : Path
        The directory where the restarts are located.
    reload : bool
        Whether to reload the restarts from the database files.
    """

    def __init__(
        self,
        directory: Union[Path, str],
        reload: bool = False,
        pickle_name: str = "experiment.pckl",
        label: str = None,
        glob_pattern: Optional[str] = None,
        progress_bar: bool = False,
    ) -> None:
        self.directory = Path(directory)
        self.database_class = Database
        self.pickle_name = pickle_name
        self._label = label
        self.glob_pattern = glob_pattern if glob_pattern is not None else "*.db"
        self.progress_bar = progress_bar
        self.uuid = uuid4()

        if reload or not (self.directory / self.pickle_name).exists():
            self.get_restarts()
        else:
            self.load()

    def get_label(self) -> str:
        if self._label is None:
            return self.directory.stem
        return self._label
    
    def get_uuid(self) -> str:
        return self.uuid

    def get_files(self, pattern: str = "*.db") -> List[Path]:
        return list(self.directory.glob(pattern))

    def get_restarts(self) -> List[RestartData]:
        if not hasattr(self, "restarts"):
            files = self.get_files(pattern=self.glob_pattern)
            self.restarts = []

            for file in track(files, disable=not self.progress_bar, transient=True, description=f"Loading: {self.get_label()}"):
                state, candidates = read_database(file, self.database_class)
                if state:
                    restart = RestartData(file, candidates)
                    self.restarts.append(restart)
            self.save()

        return self.restarts

    def get_number_of_restarts(self) -> int:
        if not hasattr(self, "restarts"):
            self.get_restarts()
        return len(self.get_restarts())

    def get_all_energies(self, fill: Union[float, int, str] = np.nan) -> np.ndarray:
        restarts = self.get_restarts()

        # Determine size of array to return:
        N_restarts = len(restarts)
        N_max = max([len(restart) for restart in restarts])

        # Get all values:
        all_values = np.zeros((N_restarts, N_max))

        for i, restart in enumerate(restarts):
            all_values[i, : len(restart)] = restart.energies.copy()
            all_values[i, len(restart) :] = fill

        return all_values

    def get_all(self, attribute: str, fill: Union[float, int, str] = np.nan) -> np.ndarray:
        restarts = self.get_restarts()

        # Determine size of array to return:
        N_restarts = len(restarts)
        N_max = max([len(restart) for restart in restarts])

        # Get all values:
        all_values = np.zeros((N_restarts, N_max))

        for i, restart in enumerate(restarts):
            all_values[i, : len(restart)] = getattr(restart, attribute).copy()
            all_values[i, len(restart) :] = fill

        return all_values

    def get_all_identifiers(self) -> List[str]:
        return [restart.identifier for restart in self.get_restarts()]

    def get_all_candidates(self) -> List[List[StandardCandidate]]:
        return [restart.candidates for restart in self.get_restarts()]

    def get_candidate(self, restart_index: int, candidate_index: int) -> StandardCandidate:
        return deepcopy(self.get_restarts()[restart_index].candidates[candidate_index])

    def get_best_candidates(self) -> List[StandardCandidate]:
        restarts = self.get_restarts()
        best_candidates = []
        for restart in restarts:
            best_index = np.argmin(restart.energies)
            best_candidates.append(restart.candidates[best_index])
        return best_candidates

    def save(self) -> None:
        try:
            with open(self.directory / self.pickle_name, "wb") as f:
                pickle.dump(self.get_restarts(), f)
        except PermissionError:
            print(f"Could not save {self.directory / self.pickle_name} due to permission error.")

    def load(self) -> None:
        with open(self.directory / self.pickle_name, "rb") as f:
            self.restarts = pickle.load(f)

    def __str__(self) -> str:
        mu = "\u03bc"
        sigma = "\u03c3"
        smin = "\u2193"
        smax = "\u2191"

        energies = self.get_all("energies")
        iterations = self.get_all("iterations")
        repr_str = f"""SearchData for: {self.directory.resolve()}"""
        repr_str += f"\n\tNumber of restarts: {self.get_number_of_restarts()}"
        repr_str += f"\n\tEnergy: {smin} = {np.nanmin(energies):.3f} {mu} = {np.nanmean(energies):.3f} {smax} = {np.nanmax(energies):.3f} {sigma} = {np.nanstd(energies):.3f}"
        repr_str += f"\n\tNumber of candidates: {len(energies[~np.isnan(energies)])}"
        repr_str += f"\n\tNumber of iterations: {np.nanmean(np.nanmax(iterations, axis=1))}"

        return repr_str


class SearchCollection:
    def __init__(
        self,
        directories: Optional[List[Union[Path, str]]] = None,
        reload: bool = False,
        labels: List[str] = None,
        progress_bar: bool = False,
    ) -> None:
        """
        Parameters
        ----------
        directories : List[Union[Path, str]]
            List of directories each containing a search.
        """
        self.reload = reload
        self._directories = []
        self._searches = []
        self.progress_bar = progress_bar

        if directories is not None:
            if labels is None:
                labels = [None for _ in directories]

            for directory, label in zip(directories, labels):
                self.add_directory(directory, label=label)

        labels = [search.get_label() for search in self._searches]

        if len(set(labels)) != len(labels):
            # Determine duplicates
            for label in set(labels):
                indices = np.argwhere(np.array(labels) == label).flatten()
                for search in [self._searches[i] for i in indices]:
                    search._label = str(search.directory.parent) + "-" + search.directory.name

    def add_directory(self, directory: Union[Path, str], label: str = None) -> None:
        """
        Add a directory to the list of directories.

        Parameters
        ----------
        directory : Union[Path, str]
            Directory containing a search.
        """
        if isinstance(directory, str):
            directory = Path(directory)
        elif isinstance(directory, Path):
            pass
        else:
            raise TypeError(f"Directory must be a Path or a string but is {type(directory)}.")
        self._directories.append(directory)
        self._searches.append(SearchData(directory, reload=self.reload, label=label, progress_bar=self.progress_bar))

    def get_searches(self) -> List[SearchData]:
        """
        Load searches from the directories.

        Returns
        -------
        List[Search]
            List of searches.
        """
        return self._searches

    def __iter__(self):  # noqa
        self.index = 0
        return self

    def __next__(self):  # noqa
        try:
            output = self._searches[self.index]
        except IndexError:
            raise StopIteration
        self.index += 1
        return output

    def __getitem__(self, index: int) -> SearchData:
        return self._searches[index]

    def __len__(self):
        return len(self._searches)



from .kmeans import KMeansSampler
from agox.utils.thermodynamics import (
    gibbs_free_energy,
    ThermodynamicsData,
)
import numpy as np
from sklearn.cluster import KMeans


class GFEKMeansSampler(KMeansSampler):
    name = "SamplerGFEKMeans"
    parameters = {}

    """
    Parameters
    ----------
    descriptor : agox.models.descriptor.ABC_descriptor.DescriptorBaseClass
        Descriptor object that inherits from DescriptorBaseClass.
    model : agox.models.model.ABC_model.ModelBaseClass
        Model object that inherits from ModelBaseClass.
    sample_size : int
        The number of sample members, or population size. 
    max_gfe : float
        The max Gibbs Free energy of formation difference, compared to the most stable structure, 
        for a structure to be considered for sampling.
    mix_stcs : bool
        Wether to preform kmeans on all of the stoichiometries together or for each stoichiometry separately, choosing the most stable one
    n_most_stable_stc : int
        If mix_stcs is False, consider for sampling n of the most stable stcs. if 0 considers all
    """

    def __init__(
        self,
        thermo_data=None,
        descriptor=None,
        model=None,
        sample_size=10,
        max_gfe=5,
        **kwargs,
    ):
        super().__init__(
            descriptor=descriptor,
            model=model,
            sample_size=sample_size,
            max_energy=max_gfe,
            **kwargs,
        )

        self.thermo_data = thermo_data

    def setup(self, all_finished_structures):
        """
        Setup up the KMeans Sampler.

        Parameters
        ----------
        all_finished_structures : list of ase.Atoms or agox.candidates.candidate.Candidate
        """

        # Check if there are any finished structures
        if len(all_finished_structures) < 1:
            self.sample = []
            return

        # Sort the structures according to gfe:
        structures, gfes = self.gfe_filter_structures(all_finished_structures)
        # Calculate features:
        X = self.get_features(structures)

        # Determine the number of clusters:
        n_clusters = 1 + min(self.sample_size - 1, int(np.floor(len(gfes) / 5)))

        # Perform KMeans clustering:
        kmeans = KMeans(
            n_clusters=n_clusters,
            init="k-means++",
            n_init=10,
            random_state=np.random.randint(0, 10e6),
        ).fit(X)
        labels = kmeans.labels_

        # For each cluster find the lowest energy member:
        sample_indices = self.select_from_clusters(gfes, n_clusters, labels)
        sample = [structures[i] for i in sample_indices]
        sample_features = [X[i] for i in sample_indices]
        cluster_centers = [kmeans.cluster_centers_[i] for i in labels[sample_indices]]

        # Sort the sample mmebers according to energy.
        # This is only important for the output.
        sample_energies = [gibbs_free_energy(candidate=t, thermo_data=self.thermo_data) for t in sample]
        sorting_indices = np.argsort(sample_energies)
        self.sample = [sample[i] for i in sorting_indices]
        self.sample_features = [sample_features[i] for i in sorting_indices]
        self.cluster_centers = [cluster_centers[i] for i in sorting_indices]
        # Print information about the sample:
        self.print_output(all_finished_structures)

    def gfe_filter_structures(self, all_structures):
        """
        Filter out structures with gfe higher than e_min + max_energy

        Parameters
        ----------
        all_structures : list of ase.Atoms or agox.candidates.candidate.Candidate

        Returns
        -------
        structures : list of ase.Atoms or agox.candidates.candidate.Candidate
            List of structures that are within the energy range.
        e : list of float
            List of energies of the structures.
        """
        # Get energies of all structures:
        gfe_all = np.array(
            [
                gibbs_free_energy(candidate=candidate, thermo_data=self.thermo_data)
                for candidate in all_structures
            ]
        )

        gfe_min = np.min(gfe_all)

        for i in range(5):
            filt = gfe_all <= gfe_min + self.max_energy * 2**i
            if np.sum(filt) >= 2 * self.sample_size:
                break
        else:
            filt = np.ones(len(gfe_all), dtype=bool)
            index_sort = np.argsort(gfe_all)
            filt[index_sort[2 * self.sample_size :]] = False

        # The final set of structures to consider:
        structures = [all_structures[i] for i in range(len(all_structures)) if filt[i]]
        e = gfe_all[filt]
        return structures, e

    def get_features(self, structures):
        features = np.array(
            [self.descriptor.get_features(s).sum(axis=0) for s in structures]
        )
        return features



from abc import ABC, abstractmethod  # noqa: N999

from agox.environments import Environment
from agox.generators import Generator
from agox.observer import Observer
from agox.samplers import Sampler


class CollectorBaseClass(ABC, Observer):
    """
    Base class for all collectors.

    Collectors produce several candidates using a set of generators.

    Parameters
    ----------
    generators : list
        List of AGOX generator instances.
    sampler : Sampler instance
        An instance of an AGOX sampler
    environment : Environment object.
        An instance of an AGOX environment object.
    order : int, optional
        Order of the collector, by default 2
    sets : dict, optional
        Dictionary of set keys, e.g. {'set_key':'candidates'}. Used to select
        in which entry in the agox.main.State cache the collector should store the candidates.
    gets : dict, optional
        Dictionary of get keys, e.g. {'get_key':'candidates'}. Used to select
        from which entry in the agox.main.State cache the collector should get candidates.
    prefix : str, optional
        Prefix for the writer, by default ''.
    surname : str, optional
        Surname for the writer, by default ''.
    """

    def __init__(
        self,
        generators: list[Generator] = None,
        sampler: Sampler = None,
        environment: Environment = None,
        order: int | float = 2,
        sets: dict[str, str] | None = None,
        gets: dict[str, str] | None = None,
        surname: str | None = None,
        **kwargs,
    ) -> None:
        """
        Quite simple, except for lines 50-54 that remove generators as observers,
        as they are now part of a Collector. This may now yield wanted behaviour
        for generators that need observers_methods that are not just for
        generation, if that becomes the case an update to those lines are
        probably necesary.

        Parameters
        ----------
        generators : list
            List of AGOX generator instances.
        sampler : Sampler instance
            An instance of an AGOX sampler
        environment : Environment object.
            An instance of an AGOX environment object.
        """
        if sets is None:
            sets = {"set_key": "candidates"}

        Observer.__init__(self, sets=sets, gets=gets, order=order, surname=surname, **kwargs)

        assert generators is not None
        assert environment is not None

        self.generators = generators
        self.sampler = sampler
        self.environment = environment
        self.candidates = []
        self.plot_confinement()

        self.add_observer_method(
            self.generate_candidates,
            sets=self.sets[0],
            gets=self.gets[0],
            order=self.order[0],
            handler_identifier="AGOX",
        )

        for generator in self.generators:
            observer_methods = [observer_method for observer_method in generator.observer_methods.values()]
            for observer_method in observer_methods:
                if observer_method.method_name == "generate":
                    generator.remove_observer_method(observer_method)

    @abstractmethod
    def make_candidates(self):  # pragma: no cover
        """
        Method that calls generators to produce a collection of candidates.

        Returns
        -------
        list
           List of candidate objects.
        """
        ...

    @abstractmethod
    def get_number_of_candidates(self, iteration):  # pragma: no cover
        """
        Method  that computes the number of candidates for the given iteration.

        Parameters
        ----------
        iteration : int
            Iteration count.

        Returns
        -------
        list
            List of integers corresponding to the self.generators list dictating
            how many candidates are generated (or really how many times each
            generator is called.)
        """
        # Number of candidates to generate.
        ...

    @Observer.observer_method
    def generate_candidates(self, state):
        # Make the candidates - this is the method that differs between versions of the class.
        if self.do_check():
            candidates = self.make_candidates()

        iteration = self.get_iteration_counter()
        for candidate in candidates:
            candidate.add_meta_information("search_iteration", iteration)

        # Add to the iteration_cache:
        state.add_to_cache(self, self.set_key, candidates, "a")
        self.writer("Number of candidates this iteration: {}".format(len(candidates)))

    ########################################################################################################################
    # Other methods
    ########################################################################################################################

    def attach(self, main):
        super().attach(main)
        for generator in self.generators:
            generator.attach(main)

    def plot_confinement(self):
        for generator in self.generators:
            generator.plot_confinement(self.environment)

    def get_fallback_generator(self):
        """
        When sampler is empty but initialized, use a RandomGenerator
        to make new candidates.
        """
        from agox.generators import RandomGenerator

        fallback_generator = RandomGenerator(**self.environment.get_confinement())
        return fallback_generator



import rich_click as click

from agox.cli.cli_analysis import cli_analysis
from agox.cli.cli_convert import cli_convert
from agox.cli.cli_graph_sorting import cli_graph_sorting
from agox.cli.cli_notebook import cli_notebook
from agox.cli.cli_plot import cli_plot


@click.group(name="agox")
@click.version_option()
def main() -> None:
    """
    Command line interface for AGOX.
    """
    return None


main.add_command(cli_convert)
main.add_command(cli_notebook)
main.add_command(cli_graph_sorting)
main.add_command(cli_analysis)
main.add_command(cli_plot)



from typing import Any, List, Optional

from rich.console import Console
from rich.panel import Panel
from rich.table import Table

from rich import style

default_writer_settings = {
    "width": 80
}

global writer_settings
writer_settings = None

global writer_verbosity
writer_verbosity = None

def get_writer_settings() -> dict:
    global writer_settings
    if writer_settings is None:
        writer_settings = default_writer_settings
    return writer_settings

def update_writer_settings(**kwargs) -> None:
    global writer_settings
    writer_settings = get_writer_settings()
    writer_settings.update(kwargs)

def get_writer_verbosity() -> int:
    global writer_verbosity
    return writer_verbosity

def set_writer_verbosity(verbosity: int) -> None:
    global writer_verbosity
    writer_verbosity = verbosity

class Writer:

    DEBUG = 1
    STANDARD = 0

    styles = {
        DEBUG: style.Style(color="dark_orange"),
        STANDARD: style.Style(color="white"),
    }
    
    def __init__(self, verbosity: int = STANDARD, **kwargs) -> None:
        self.console_kwargs = get_writer_settings()
        self.console_kwargs.update(kwargs)
        self._verbosity = verbosity


    def writer(self, string: str, *args, verbosity: int = STANDARD, **kwargs) -> None:
        if verbosity <= self.verbosity:
            console = Console(**self.console_kwargs)
            console.print(string, *args, **kwargs, style=self.styles[verbosity])    

    def __call__(self, string: str, *args, **kwargs) -> None:
        self.writer(string, *args, verbosity=self.STANDARD, **kwargs)

    def debug(self, string: str, *args, **kwargs) -> None:
        self.writer(string, *args, verbosity=self.DEBUG, **kwargs)

    def write_header(self, string: str) -> None:
        console = Console(**self.console_kwargs)
        console.rule(string)

    def write_table(self, table_columns: List, table_rows: List, **table_kwargs) -> None:
        console = Console(**self.console_kwargs)

        table = Table(**table_kwargs)
        for column in table_columns:
            table.add_column(column)

        for row in table_rows:
            table.add_row(*row)

        console.print(table)

    def write_panel(self, panel_content: str, panel_title: Optional[str] = None) -> None:
        console = Console(**self.console_kwargs)
        panel = Panel(panel_content, title=panel_title)
        console.print(panel)

    @staticmethod
    def update_settings(**kwargs) -> None:
        update_writer_settings(**kwargs)

    @property
    def verbosity(self) -> int:
        global_writer_verbosity = get_writer_verbosity()

        if global_writer_verbosity is None:
            verbosity = self._verbosity
        elif global_writer_verbosity > self._verbosity:
            verbosity = global_writer_verbosity
        else:
            verbosity = self._verbosity

        return verbosity


    


import numpy as np
from ase.calculators.singlepoint import SinglePointCalculator

from agox.candidates import Candidate
from agox.helpers.confinement import Confinement
from agox.postprocessors.ABC_postprocess import PostprocessBaseClass


class CenteringPostProcess(PostprocessBaseClass):
    """
    Centers a candidate object to the middle of the cell.

    Should not be used for periodic systems.
    """

    name = "CenteringPostProcess"

    def __init__(self, confinement: Confinement | None = None, **kwargs) -> None:
        super().__init__(**kwargs)
        self.confinement = confinement

    def postprocess(self, candidate: Candidate) -> Candidate:
        """
        Centers a candidate object to the middle of the cell.

        Parameters
        ----------
        candidate : Candidate
            Candidate object to be centered.

        Returns
        -------
        candidate : Candidate
            Centered candidate object.

        """
        center_candidate = candidate.copy()
        if candidate.pbc.any():
            self.writer("CenteringPostProcess used on a periodic system, won't have any effect. Consider removing from script.")
            return candidate

        results = None
        if hasattr(candidate, "calc"):
            if hasattr(candidate.calc, "results"): 
                results = candidate.calc.results
                    
        com = candidate.get_center_of_mass()
        cell_middle = np.sum(candidate.get_cell(), 0) / 2
        center_candidate.positions = candidate.positions - com + cell_middle

        if self.confinement is not None:
            if not self.confinement.check_confinement(center_candidate.positions).all():
                self.writer.debug('Confinement violated after centering.')
                return candidate
        
        if results is not None: 
            center_candidate.calc = SinglePointCalculator(center_candidate, **results)
        return center_candidate



import time
from abc import ABC, abstractmethod
from typing import Callable

from agox import State
from agox.candidates import Candidate, CandidateBaseClass
from agox.observer import Observer


class EvaluatorBaseClass(ABC, Observer):
    """
    Base class for evaluators.

    Evaluators calculate the objective function (usually the energy).

    Parameters
    ----------
    number_to_evaluate : int, optional
        The number of candidates to evaluate before stopping. The default is 1.
    check_callback : Callable[[CandidateBaseClass], None], optional
        A function that is called with every evaluated candidate as argument
        before it is added to the list of evaluated candidates. If this
        callable raises an exception, the candidate will not be added. This can
        be used to, e.g., check for convergence when using ASE calculators that
        do not raise an exception themselves if they did not converge. When not
        provided, no check will be performed.
    """

    def __init__(
        self,
        number_to_evaluate: int = 1,
        max_evaluate: int = None,
        check_callback: Callable[[CandidateBaseClass], None] | None = None,
        gets: dict[str, str] | None = None,
        sets: dict[str, str] | None = None,
        order: int | float = 5,
        surname: str | None = None,
        **kwargs,
    ) -> None:

        if sets is None:
            sets = {"set_key": "evaluated_candidates"}
        if gets is None:
            gets = {"get_key": "prioritized_candidates"}

        Observer.__init__(self, gets=gets, sets=sets, order=order, surname=surname, **kwargs)
        self.number_to_evaluate = number_to_evaluate
        self.check_callback = check_callback or self._default_check_callback
        self.total_evaluated = 0
        self.max_evaluate = max_evaluate

        self.evaluated_candidates = []

        self.add_observer_method(
            self.evaluate, sets=self.sets[0], gets=self.gets[0], order=self.order[0], handler_identifier="AGOX"
        )

        self.start_time = time.time()

    def __call__(self, candidate):
        return self.evaluate_candidate(candidate)

    @abstractmethod
    def evaluate_candidate(self, candidate: list[Candidate]) -> None:  # pragma: no cover
        """
        Evaluates the given candidate.

        This function MUST append the candidates it wishes to pass to the AGOX
        State cache to the list self.evaluated_candidates.

        Parameters
        ----------
        candidate : AGOX candidate object.
            The candidate object to evaluate.

        Returns
        -------
        bool
            Whether the evaluation was successful.
        """
        ...

    @property
    @abstractmethod
    def name(self) -> None:  # pragma: no cover
        pass

    @Observer.observer_method
    def evaluate(self, state: State) -> None:
        candidates = state.get_from_cache(self, self.get_key)
        done = False

        self.evaluated_candidates = []
        passed_evaluation_count = 0
        if self.do_check():
            while candidates and not done:
                self.writer(f"Trying candidate - remaining {len(candidates)}")
                candidate = candidates.pop(0)

                if candidate is None:
                    self.writer("Candidate was None - are your other modules working as intended?")
                    continue

                internal_state = self.evaluate_candidate(candidate)

                if internal_state:
                    self.writer("Succesful calculation of candidate.")
                    passed_evaluation_count += 1
                    self.evaluated_candidates[-1].add_meta_information("final", True)
                    iteration = self.get_iteration_counter()
                    energy_dict = {iteration : self.evaluated_candidates[-1].get_potential_energy()}
                    self.evaluated_candidates[-1].add_meta_information("evaluator_energy", energy_dict)
                    if passed_evaluation_count == self.number_to_evaluate:
                        self.writer("Calculated required number of candidates.")
                        done = True

        state.add_to_cache(self, self.set_key, self.evaluated_candidates, mode="a")

        self.total_evaluated += len(self.evaluated_candidates)
        if self.max_evaluate is not None and self.total_evaluated >= self.max_evaluate:
            state.set_convergence_status(True)

    def _check_and_append_candidate(self, candidate: CandidateBaseClass) -> None:
        """Check a candidate using the check callable, and then append it to
        the list of evaluated candidates.

        Parameters
        ----------
        candidate : CandidateBaseClass
            Candidate to check and append.
        """
        self.check_callback(candidate)
        self._add_time_stamp(candidate)
        self.evaluated_candidates.append(candidate)

    def _default_check_callback(self, _: CandidateBaseClass) -> None:
        """Default check callable: do nothing."""
        pass
        
    def _add_time_stamp(self, candidate):
        candidate.add_meta_information("time_stamp", time.time() - self.start_time)

    def __add__(self, other):
        return EvaluatorCollection(evaluators=[self, other])


class EvaluatorCollection(EvaluatorBaseClass):
    name = "EvaluatorCollection"

    def __init__(self, evaluators: list[EvaluatorBaseClass]) -> None:
        super().__init__()
        self.evaluators = evaluators

    def evaluate_candidate(self, candidate: Candidate) -> bool:
        self.apply_evaluators(candidate)

    def add_evaluator(self, evaluator: EvaluatorBaseClass) -> None:
        self.evaluators.append(evaluator)

    def list_evaluators(self) -> None:
        for i, evaluator in enumerate(self.evaluators):
            print("Evaluator {}: {} - {}".format(i, evaluator.name, evaluator))

    def apply_evaluators(self, candidate: Candidate) -> bool:
        for evaluator in self.evaluators:
            evaluator_state = evaluator(candidate)
            if not evaluator_state:
                return False
        return True

    def __add__(self, other: EvaluatorBaseClass): # noqa
        self.evaluators.append(other)
        return self



from abc import ABC, abstractmethod  # noqa: N999
from typing import List

from ase import Atoms
from ase.calculators.singlepoint import SinglePointCalculator

from agox.candidates import Candidate
from agox.main.state import State
from agox.observer import Observer, ObserverHandler


class DatabaseBaseClass(ABC, ObserverHandler, Observer):
    """
    Base class for all databases.

    Databases are used to store, typically evaluated, candidates.

    Parameters
    ----------
    gets : dict
        Dictionary of get keys, e.g. {'get_key':'evaluated_candidates'}. Used to select
        from which entry in the agox.main.State cache the database should get candidates.
    order : int
        Order of the database, by default 6
    """

    def __init__(
        self,
        gets: dict[str, str] | None = None,
        sets: dict[str, str] | None = None,
        order: int = 6,
        surname: str = "",
        **kwargs,
    ) -> None:

        if gets is None:
            gets = {"get_key": "evaluated_candidates"}        

        Observer.__init__(self, gets=gets, sets=sets, order=order, surname=surname, **kwargs)
        ObserverHandler.__init__(self, handler_identifier="database", dispatch_method=self.store_in_database)
        self.candidates = []

        self.objects_to_assign = []

        self.add_observer_method(
            self.store_in_database, sets=self.sets[0], gets=self.gets[0], order=self.order[0], handler_identifier="AGOX"
        )

    ########################################################################################
    # Required methods
    ########################################################################################

    @abstractmethod
    def write(self, *args, **kwargs):  # pragma: no cover # noqa
        """
        Write stuff to database
        """

    @abstractmethod
    def store_candidate(self, candidate):  # pragma: no cover # noqa
        pass

    @property
    @abstractmethod
    def name(self):  # pragma: no cover # noqa
        return NotImplementedError

    ########################################################################################
    # Default methods
    ########################################################################################

    def _copy_candidate(self, candidate: Candidate) -> Candidate:
        new_candidate = candidate.copy()
        calc_dict = candidate.calc.results.copy()
        new_calc = SinglePointCalculator(candidate, **calc_dict)
        new_candidate.set_calculator(new_calc)
        return new_candidate

    def get_all_candidates(self, **kwargs) -> List[Candidate]:
        all_candidates = []
        for candidate in self.candidates:
            all_candidates.append(self._copy_candidate(candidate))
        return all_candidates

    def get_most_recent_candidate(self) -> Candidate:
        if len(self.candidates) > 0:
            candidate = self._copy_candidate(self.candidates[-1])
        else:
            candidate = None
        return candidate

    def get_recent_candidates(self, number: int) -> List[Candidate]:
        return [self._copy_candidate(candidate) for candidate in self.candidates[-number:]]

    def __len__(self) -> int:
        return len(self.candidates)

    @Observer.observer_method
    def store_in_database(self, state: State) -> None:
        evaluated_candidates = state.get_from_cache(self, self.get_key)
        anything_accepted = False
        for j, candidate in enumerate(evaluated_candidates):
            if candidate:
                self.store_candidate(candidate, accepted=True, write=True)
                anything_accepted = True

            elif candidate is None:
                dummy_candidate = self.candidate_instanstiator(template=Atoms())
                dummy_candidate.set_calculator(SinglePointCalculator(dummy_candidate, energy=float("nan")))

                # This will dispatch to observers if valid data has been added but the last candidate is None.
                self.store_candidate(candidate, accepted=False, write=True)

        if anything_accepted:
            self.dispatch_to_observers(database=self, state=state)



from .ABC_sparsifier import SparsifierBaseClass
from .CUR import CUR
from .MBkmeans import MBkmeans
from .random import RandomSparsifier

__all__ = [
    "SparsifierBaseClass",
    "CUR",
    "RandomSparsifier",
    "MBkmeans",
]
